import argparse
import os
import sys

import urllib3
from docker.client import DockerClient
from docker.types import EndpointSpec, RestartPolicy

urllib3.disable_warnings()

# Set env to False if the container should be left around for debug purposes, etc.
remove_container = bool(
    os.getenv("REMOVE_CONTAINER", os.getenv("GP_REMOVE_CONTAINER", "True")).lower() == "true"
)
swarm_mode = bool(os.getenv("DOCKER_MODE", os.getenv("GP_DOCKER_MODE", "swarm")).lower() == "swarm")


def launch_docker_kernel(
    kernel_id,
    port_range,
    response_addr,
    public_key,
    spark_context_init_mode,
    kernel_class_name,
):
    # Launches a containerized kernel.

    # Can't proceed if no image was specified.
    image_name = os.environ.get("KERNEL_IMAGE", None)
    if image_name is None:
        sys.exit("ERROR - KERNEL_IMAGE not found in environment - kernel launch terminating!")

    # Container name is composed of KERNEL_USERNAME and KERNEL_ID
    container_name = os.environ.get("KERNEL_USERNAME", "") + "-" + kernel_id

    # Determine network. If EG_DOCKER_NETWORK has not been propagated, fall back to 'bridge'...
    docker_network = os.environ.get("DOCKER_NETWORK", os.environ.get("EG_DOCKER_NETWORK", "bridge"))

    # Build labels - these will be modelled similar to kubernetes: kernel_id, component, app, ...
    labels = {}
    labels["kernel_id"] = kernel_id
    labels["component"] = "kernel"
    labels["app"] = "gateway-provisioners"

    # Capture env parameters...
    param_env = {}
    param_env["PORT_RANGE"] = port_range
    param_env["PUBLIC_KEY"] = public_key
    param_env["RESPONSE_ADDRESS"] = response_addr
    param_env["KERNEL_SPARK_CONTEXT_INIT_MODE"] = spark_context_init_mode
    if kernel_class_name:
        param_env["KERNEL_CLASS_NAME"] = kernel_class_name

    # Since the environment is specific to the kernel (per env stanza of kernelspec, KERNEL_ and EG_CLIENT_ENVS)
    # just add the env here.
    param_env.update(os.environ)
    param_env.pop(
        "PATH"
    )  # Let the image PATH be used.  Since this is relative to images, we're probably safe.

    user = param_env.get("KERNEL_UID")
    group = param_env.get("KERNEL_GID")

    # setup common args
    kwargs = {}
    kwargs["name"] = container_name
    kwargs["hostname"] = container_name
    kwargs["user"] = user
    kwargs["labels"] = labels

    client = DockerClient.from_env()
    if swarm_mode:
        networks = []
        networks.append(docker_network)
        # mounts = list()  # Enable if necessary
        # mounts.append("/usr/local/share/jupyter/kernels:/usr/local/share/jupyter/kernels:ro")
        endpoint_spec = EndpointSpec(mode="dnsrr")
        restart_policy = RestartPolicy(condition="none")

        # finish args setup
        kwargs["env"] = param_env
        kwargs["endpoint_spec"] = endpoint_spec
        kwargs["restart_policy"] = restart_policy
        kwargs["container_labels"] = labels
        kwargs["networks"] = networks
        kwargs["groups"] = [group, "100"]
        if param_env.get("KERNEL_WORKING_DIR"):
            kwargs["workdir"] = param_env.get("KERNEL_WORKING_DIR")
        # kwargs['mounts'] = mounts   # Enable if necessary
        # print("service args: {}".format(kwargs))  # useful for debug
        client.services.create(image_name, **kwargs)
    else:
        # volumes = {  # Enable if necessary
        #     "/usr/local/share/jupyter/kernels": {
        #         "bind": "/usr/local/share/jupyter/kernels",
        #         "mode": "ro",
        #     }
        # }

        # finish args setup
        kwargs["environment"] = param_env
        kwargs["remove"] = remove_container
        kwargs["network"] = docker_network
        kwargs["group_add"] = [group, "100"]
        kwargs["detach"] = True
        if param_env.get("KERNEL_WORKING_DIR"):
            kwargs["working_dir"] = param_env.get("KERNEL_WORKING_DIR")
        # kwargs['volumes'] = volumes   # Enable if necessary
        # print("container args: {}".format(kwargs))  # useful for debug
        client.containers.run(image_name, **kwargs)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--kernel-id",
        dest="kernel_id",
        nargs="?",
        help="Indicates the id associated with the launched kernel.",
    )
    parser.add_argument(
        "--port-range",
        dest="port_range",
        nargs="?",
        metavar="<lowerPort>..<upperPort>",
        help="Port range to impose for kernel ports",
    )
    parser.add_argument(
        "--response-address",
        dest="response_address",
        nargs="?",
        metavar="<ip>:<port>",
        help="Connection address (<ip>:<port>) for returning connection file",
    )
    parser.add_argument(
        "--public-key",
        dest="public_key",
        nargs="?",
        help="Public key used to encrypt connection information",
    )
    parser.add_argument(
        "--spark-context-initialization-mode",
        dest="spark_context_init_mode",
        nargs="?",
        help="Indicates whether or how a spark context should be created",
        default="none",
    )
    parser.add_argument(
        "--kernel-class-name",
        dest="kernel_class_name",
        nargs="?",
        help="Indicates the name of the kernel class to use.  Must be a subclass of 'ipykernel.kernelbase.Kernel'.",
    )
    arguments = vars(parser.parse_args())
    kernel_id = arguments["kernel_id"]
    port_range = arguments["port_range"]
    response_addr = arguments["response_address"]
    public_key = arguments["public_key"]
    spark_context_init_mode = arguments["spark_context_init_mode"]
    kernel_class_name = arguments["kernel_class_name"]

    launch_docker_kernel(
        kernel_id,
        port_range,
        response_addr,
        public_key,
        spark_context_init_mode,
        kernel_class_name,
    )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import os
import sys

from jupyter_core.application import JupyterApp
from overrides import overrides
from traitlets import Bool, Unicode, default

from .._version import __version__
from .base_app import DEFAULT_LANGUAGE, PYTHON, SCALA, BaseSpecSparkApp, R

DASK = "dask"

DEFAULT_KERNEL_NAMES = {
    PYTHON: "yarn_spark_python",
    SCALA: "yarn_spark_scala",
    R: "yarn_spark_r",
    DASK: "yarn_dask_python",
}
DEFAULT_DISPLAY_NAMES = {
    PYTHON: "Spark Python (YARN Cluster)",
    SCALA: "Spark Scala (YARN Cluster)",
    R: "Spark R (YARN Cluster)",
    DASK: "Dask Python (YARN Cluster)",
}


class YarnSpecInstaller(BaseSpecSparkApp):
    """CLI for extension management."""

    name = "jupyter-yarn-spec"
    description = "Creates a Jupyter kernel specification for use within a Hadoop Yarn cluster."
    # Note that the left justification of the second example is necessary to ensure proper
    # alignment with the first example during --help output.
    examples = """
    jupyter-yarn-spec install --language=R --spark-home=/usr/local/spark

jupyter-yarn-spec install --kernel-name=dask_python --dask --yarn-endpoint=http://foo.bar:8088/ws/v1/cluster

jupyter-yarn-spec install --language=Scala --spark-init-mode eager
    """

    @default("kernel_name")
    def _kernel_name_default(self) -> str:
        return DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]

    @default("display_name")
    def _display_name_default(self) -> str:
        return DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]

    # Yarn endpoint
    yarn_endpoint_env = "GP_YARN_ENDPOINT"
    yarn_endpoint = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The http url specifying the YARN Resource Manager. Note: If this value is NOT set,
the YARN library will use the files within the local HADOOP_CONFIG_DIR to determine the
active resource manager. (GP_YARN_ENDPOINT env var)""",
    )

    @default("yarn_endpoint")
    def _yarn_endpoint_default(self):
        return os.getenv(self.yarn_endpoint_env)

    # Alt Yarn endpoint
    alt_yarn_endpoint_env = "GP_ALT_YARN_ENDPOINT"
    alt_yarn_endpoint = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The http url specifying the alternate YARN Resource Manager.  This value should
be set when YARN Resource Managers are configured for high availability.  Note: If both
YARN endpoints are NOT set, the YARN library will use the files within the local
HADOOP_CONFIG_DIR to determine the active resource manager.
(GP_ALT_YARN_ENDPOINT env var)""",
    )

    @default("alt_yarn_endpoint")
    def _alt_yarn_endpoint_default(self):
        return os.getenv(self.alt_yarn_endpoint_env)

    python_root = Unicode(
        sys.prefix,
        config=True,
        help="Specify where the root of the python installation resides (parent dir of bin/python).",
    )

    extra_dask_opts = Unicode("", config=True, help="Specify additional Dask options.")

    # Flags
    # Yarn endpoint security enabled
    yarn_endpoint_security_enabled_env = "GP_YARN_ENDPOINT_SECURITY_ENABLED"
    yarn_endpoint_security_enabled_default_value = False
    yarn_endpoint_security_enabled = Bool(
        yarn_endpoint_security_enabled_default_value,
        config=True,
        help="""Is YARN Kerberos/SPNEGO Security enabled (True/False).
(GP_YARN_ENDPOINT_SECURITY_ENABLED env var)""",
    )

    @default("yarn_endpoint_security_enabled")
    def _yarn_endpoint_security_enabled_default(self):
        return bool(
            os.getenv(
                self.yarn_endpoint_security_enabled_env,
                self.yarn_endpoint_security_enabled_default_value,
            )
        )

    # Impersonation enabled
    impersonation_enabled_env = "GP_IMPERSONATION_ENABLED"
    impersonation_enabled = Bool(
        False,
        config=True,
        help="""Indicates whether impersonation will be performed during kernel launch.
(GP_IMPERSONATION_ENABLED env var)""",
    )

    @default("impersonation_enabled")
    def _impersonation_enabled_default(self):
        return bool(os.getenv(self.impersonation_enabled_env, "false").lower() == "true")

    dask = Bool(False, config=True, help="Kernelspec will be configured for Dask YARN.")

    aliases = {
        "yarn-endpoint": "YarnSpecInstaller.yarn_endpoint",
        "alt-yarn-endpoint": "YarnSpecInstaller.alt_yarn_endpoint",
        "python-root": "YarnSpecInstaller.python_root",
        "extra-dask-opts": "YarnSpecInstaller.extra_dask_opts",
    }
    aliases.update(BaseSpecSparkApp.aliases)

    flags = {
        "dask": (
            {"YarnSpecInstaller": {"dask": True}},
            dask.help,
        ),
        "yarn-endpoint-security-enabled": (
            {"YarnSpecInstaller": {"yarn_endpoint_security_enabled": True}},
            yarn_endpoint_security_enabled.help,
        ),
        "impersonation-enabled": (
            {"YarnSpecInstaller": {"impersonation_enabled": True}},
            impersonation_enabled.help,
        ),
    }
    flags.update(BaseSpecSparkApp.flags)

    @overrides
    def detect_missing_extras(self):
        super().detect_missing_extras()
        try:
            import yarn_api_client  # noqa: F401
        except ImportError:
            self.log.warning(
                "The extra package 'yarn_api_client'is not installed in this environment and is "
                "required.  Ensure that gateway_provisioners is installed by specifying the "
                "extra 'yarn' (e.g., pip install 'gateway_provisioners[yarn]')."
            )

    @overrides
    def validate_parameters(self):
        super().validate_parameters()

        entered_language = self.language
        self.language = self.language.lower()
        self.launcher_dir_name = self.language
        self.resource_dir_name = self.language

        if self.dask:
            if self.language != PYTHON:
                self.log.warning(
                    "Dask support only works with Python, changing language from {} to Python.".format(
                        entered_language
                    )
                )
                self.language = PYTHON
            # if kernel and display names are still defaulted, silently change to dask defaults
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[DASK]
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[DASK]

            self.kernel_spec_dir_name = DEFAULT_KERNEL_NAMES[DASK]
            self.spark_init_mode = "none"
            if len(self.extra_spark_opts) > 0:
                self.log.warning("--extra_spark_opts will be ignored for Dask-based kernelspecs.")
                self.extra_spark_opts = ""
        else:
            # if kernel and display names are still defaulted, silently change to language defaults
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[self.language]
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[self.language]

            self.kernel_spec_dir_name = DEFAULT_KERNEL_NAMES[self.language]
            if len(self.extra_dask_opts) > 0:
                self.log.warning("--extra_dask_opts will be ignored for Spark-based kernelspecs.")
                self.extra_dask_opts = ""

        # sanitize kernel_name
        self.kernel_name = self.kernel_name.replace(" ", "_")

    @overrides
    def add_optional_config_entries(self, config_stanza: dict) -> None:
        super().add_optional_config_entries(config_stanza)
        if self.yarn_endpoint and self.yarn_endpoint != self.yarn_endpoint_default():
            config_stanza["yarn_endpoint"] = self.yarn_endpoint
        if self.alt_yarn_endpoint and self.alt_yarn_endpoint != self.alt_yarn_endpoint_default():
            config_stanza["alt_yarn_endpoint"] = self.alt_yarn_endpoint
        if (
            self.yarn_endpoint_security_enabled
            and self.yarn_endpoint_security_enabled != self.yarn_endpoint_security_enabled_default()
        ):
            config_stanza["yarn_endpoint_security_enabled"] = self.yarn_endpoint_security_enabled

    @overrides
    def get_substitutions(self, install_dir) -> dict:
        substitutions = super().get_substitutions(install_dir)
        substitutions["extra_dask_opts"] = self.extra_dask_opts
        substitutions["python_root"] = self.python_root
        substitutions["py4j_path"] = ""

        # If this is a python kernel, attempt to get the path to the py4j file.
        if self.language == PYTHON and not self.dask:
            try:
                python_lib_contents = os.listdir(f"{self.spark_home}/python/lib")
                py4j_zip = list(filter(lambda filename: "py4j" in filename, python_lib_contents))[0]

                # This is always a sub-element of a path, so let's prefix with colon
                substitutions["py4j_path"] = f":{self.spark_home}/python/lib/{py4j_zip}"
            except OSError:
                self.log.warn("Unable to find py4j, installing without PySpark support.")
        return substitutions


class YarnProvisionerApp(JupyterApp):
    """Application responsible for driving the creation of Yarn-based kernel specifications."""

    version = __version__
    name = "jupyter yarn-spec"
    description = """Application used to create kernel specifications for use on Hadoop Yarn clusters
    via the YarnProvisioner kernel provisioner."""
    subcommands = dict(
        {
            "install": (
                YarnSpecInstaller,
                YarnSpecInstaller.description.splitlines()[0],
            ),
        }
    )
    aliases = {}
    flags = {}

    def start(self):
        super().start()

        if self.subapp is None:
            print(f"No subcommand specified. Must specify one of: {list(self.subcommands)}")
            print()
            self.print_description()
            self.print_subcommands()
            self.exit(1)


if __name__ == "__main__":
    YarnProvisionerApp.launch_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import base64
import json
import logging
import os
import random
import signal
import socket
import uuid
from multiprocessing import Process, set_start_method
from typing import Any, Optional

from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.PublicKey import RSA
from Cryptodome.Random import get_random_bytes
from Cryptodome.Util.Padding import pad
from jupyter_client.connect import write_connection_file

LAUNCHER_VERSION = 1  # Indicate to server the version of this launcher (payloads may vary)

max_port_range_retries = int(os.getenv("MAX_PORT_RANGE_RETRIES", "5"))

log_level = os.getenv("LOG_LEVEL", "10")
log_level = int(log_level) if log_level.isdigit() else log_level

logging.basicConfig(format="[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s] %(message)s")

logger = logging.getLogger("server_listener")
logger.setLevel(log_level)

# Global that can be set to True in signal handler (SIGTERM)
# when parent kernel process is asked to shutdown.
shutdown: bool = False


class ServerListener:
    def __init__(
        self,
        conn_filename: str,
        parent_pid: int,
        lower_port: int,
        upper_port: int,
        response_addr: str,
        kernel_id: str,
        public_key: str,
        cluster_type: Optional[str] = None,
    ):
        # Note, in the R integration, parameters come into Python as strings, so
        # we need to explicitly cast non-strings.
        self.conn_filename: str = conn_filename
        self.parent_pid: int = int(parent_pid)
        self.lower_port: int = int(lower_port)
        self.upper_port: int = int(upper_port)
        self.response_addr: str = response_addr
        self.kernel_id: str = kernel_id
        self.public_key: bytes = public_key.encode("utf-8")
        self.cluster_type: str = cluster_type

        # Initialized later...
        self.comm_socket: socket | None = None

    def build_connection_file(self) -> None:
        ports: list = self._select_ports(5)
        write_connection_file(
            fname=self.conn_filename,
            ip="0.0.0.0",  # noqa: S104
            key=str(uuid.uuid4()).encode("utf-8"),  # convert to bytes,
            shell_port=ports[0],
            iopub_port=ports[1],
            stdin_port=ports[2],
            hb_port=ports[3],
            control_port=ports[4],
        )

    def _encrypt(self, connection_info_bytes: bytes) -> bytes:
        """Encrypt the connection information using a generated AES key that is then encrypted using
        the public key passed from the server.  Both are then returned in an encoded JSON payload.
        """
        aes_key = get_random_bytes(16)
        cipher = AES.new(aes_key, mode=AES.MODE_ECB)

        # Encrypt the connection info using the aes_key
        encrypted_connection_info = cipher.encrypt(pad(connection_info_bytes, 16))
        b64_connection_info = base64.b64encode(encrypted_connection_info)

        # Encrypt the aes_key using the server's public key
        imported_public_key = RSA.importKey(base64.b64decode(self.public_key))
        cipher = PKCS1_v1_5.new(key=imported_public_key)
        encrypted_key = base64.b64encode(cipher.encrypt(aes_key))

        # Compose the payload and Base64 encode it
        payload = {
            "version": LAUNCHER_VERSION,
            "key": encrypted_key.decode(),
            "conn_info": b64_connection_info.decode(),
        }
        b64_payload = base64.b64encode(json.dumps(payload).encode(encoding="utf-8"))
        return b64_payload

    def return_connection_info(self) -> None:
        """Returns the connection information corresponding to this kernel."""
        response_parts = self.response_addr.split(":")
        if len(response_parts) != 2:
            logger.error(
                f"Invalid format for response address '{self.response_addr}'. Assuming 'pull' mode..."
            )
            return

        response_ip = response_parts[0]
        try:
            response_port = int(response_parts[1])
        except ValueError:
            logger.error(
                f"Invalid port component found in response address '{self.response_addr}'. Assuming 'pull' mode..."
            )
            return

        with open(self.conn_filename) as fp:
            cf_json = json.load(fp)
            fp.close()

        # add process and process group ids into connection info
        cf_json["pid"] = self.parent_pid
        cf_json["pgid"] = os.getpgid(self.parent_pid)

        # prepare socket address for handling signals
        self.prepare_comm_socket()  # self.comm_socket initialized
        cf_json["comm_port"] = self.comm_socket.getsockname()[1]
        cf_json["kernel_id"] = self.kernel_id

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((response_ip, response_port))
            json_content = json.dumps(cf_json).encode(encoding="utf-8")
            logger.debug(f"JSON Payload '{json_content}")
            payload = self._encrypt(json_content)
            logger.debug(f"Encrypted Payload '{payload}")
            s.send(payload)

    def prepare_comm_socket(self) -> None:
        """Prepares the socket to which the server will send signal and shutdown requests."""
        self.comm_socket = self._select_socket()
        logger.info(
            f"Signal socket bound to host: "
            f"{self.comm_socket.getsockname()[0]}, port: {self.comm_socket.getsockname()[1]}"
        )
        self.comm_socket.listen(1)
        self.comm_socket.settimeout(5)

    def _select_ports(self, count: int) -> list:
        """Select and return n random ports that are available and adhere to the given port range, if applicable."""
        ports = []
        sockets = []
        for _ in range(count):
            sock = self._select_socket()
            ports.append(sock.getsockname()[1])
            sockets.append(sock)
        for sock in sockets:
            sock.close()
        return ports

    def _select_socket(self) -> socket:
        """Create and return a socket whose port is available and adheres to the given port range, if applicable."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        found_port = False
        retries = 0
        while not found_port:
            try:
                sock.bind(("0.0.0.0", self._get_candidate_port()))  # noqa: S104
                found_port = True
            except OSError as ose:
                retries = retries + 1
                if retries > max_port_range_retries:
                    err_msg = (
                        f"Failed to locate port within range {self.lower_port}..{self.upper_port} "
                        f"after {max_port_range_retries} retries!"
                    )
                    raise RuntimeError(err_msg) from ose
        return sock

    def _get_candidate_port(self) -> int:
        """Returns a port within the given range.  If the range is zero, the zero is returned."""
        range_size = self.upper_port - self.lower_port
        if range_size == 0:
            return 0
        return random.randint(self.lower_port, self.upper_port)

    def get_server_request(self) -> dict:
        """Gets a request from the server and returns the corresponding dictionary."""
        conn: socket = None
        data: str = ""
        request_info: Optional[dict] = None
        try:
            conn, addr = self.comm_socket.accept()
            while True:
                buffer: bytes = conn.recv(1024)
                if buffer == b"":  # send is complete
                    if len(data) > 0:
                        request_info = json.loads(data)
                    else:
                        logger.info("DEBUG: get_server_request: no data received - returning None")
                    break
                data = data + buffer.decode(
                    "utf-8"
                )  # append what we received until we get no more...
        except Exception as ex:
            if type(ex) is not socket.timeout:
                raise ex
        finally:
            if conn:
                conn.close()

        return request_info

    def process_requests(self) -> None:
        """Waits for requests from the server and processes each when received.  Currently,
        these will be one of a sending a signal to the corresponding kernel process (signum) or
        stopping the listener and exiting the kernel (shutdown).
        """
        global shutdown

        # Setup signal handler for SIGTERM so we can detect that kernel process is
        # terminating its children (IPyKernel does this).
        signal.signal(signal.SIGTERM, handle_sigterm)

        # Since this creates the communication socket, we should do this here so the socket
        # gets created in the sub-process/thread.  This is necessary on MacOS/Python.
        self.return_connection_info()

        while not shutdown:
            request = self.get_server_request()
            if request:
                signum = -1  # prevent logging poll requests since that occurs every 3 seconds
                if request.get("signum") is not None:
                    signum = int(request.get("signum"))
                    os.kill(self.parent_pid, signum)
                    if signum == 2 and self.cluster_type == "spark":
                        os.kill(self.parent_pid, signal.SIGUSR2)
                if request.get("shutdown") is not None:
                    shutdown = bool(request.get("shutdown"))
                if signum != 0:
                    logger.info(f"server_listener got request: {request}")

        logger.info("ServerListener.process_requests exiting.")


def handle_sigterm(sig: int, frame: Any) -> None:
    """Revert to the default handler and set shutdown to True."""
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    logger.info("SIGTERM caught and reset to default handler...")
    global shutdown
    shutdown = True


def setup_server_listener(
    conn_filename: str,
    parent_pid: int,
    lower_port: int,
    upper_port: int,
    response_addr: str,
    kernel_id: str,
    public_key: str,
    cluster_type: Optional[str] = None,
) -> None:
    """Initializes the server listener sub-process to handle requests from the server."""

    # Create the ServerListener instance and build the connection file PRIOR to sub-process.
    # This is synchronous relative to the launcher, so the launcher can start the kernel
    # using the connection file and no race condition is introduced.
    sl = ServerListener(
        conn_filename,
        parent_pid,
        lower_port,
        upper_port,
        response_addr,
        kernel_id,
        public_key,
        cluster_type,
    )
    sl.build_connection_file()

    set_start_method("fork")
    server_listener = Process(target=sl.process_requests)
    server_listener.start()


__all__ = [
    "setup_server_listener",
]



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
from uuid import uuid4

import pytest
from jupyter_client import KernelConnectionInfo
from validators import TEST_USER, K8sValidator


@pytest.mark.parametrize("seed_env", [{"KERNEL_USERNAME": TEST_USER}])
async def test_lifecycle(init_api_mocks, response_manager, get_provisioner, seed_env):
    name = "kubernetes"
    kernel_id = str(uuid4())
    validator = K8sValidator.create_instance(
        name, seed_env, kernel_id=kernel_id, response_manager=response_manager
    )
    os.environ.update(seed_env)

    provisioner = get_provisioner(name, kernel_id)
    validator.validate_provisioner(provisioner)

    kwargs = {"env": seed_env}
    kwargs = await provisioner.pre_launch(**kwargs)
    validator.validate_pre_launch(kwargs)

    cmd = kwargs.pop("cmd")
    connection_info: KernelConnectionInfo = await provisioner.launch_kernel(cmd, **kwargs)
    validator.validate_launch_kernel(connection_info)

    await provisioner.post_launch(**kwargs)
    validator.validate_post_launch(kwargs)

    assert provisioner.has_process is True, "has_process property has unexpected value: False"

    poll_result = await provisioner.poll()
    assert poll_result is None, f"poll() returned unexpected result: '{poll_result}'"

    # send_signal()? only tests remote provisioner and probably better-suited for launcher tests

    # In the container-based provisioners, kill and terminate are identical, so only testing terminate.

    await provisioner.terminate()

    # shutdown_requested() would only test remote provisioner and probably better-suited for launcher tests

    await provisioner.cleanup(restart=False)
    assert provisioner.has_process is False, "has_process property has unexpected value: True"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from subprocess import Popen

from .docker_client import DockerResource, docker_resources
from .k8s_client import K8sResource, k8s_resources
from .yarn_client import YarnResource, yarn_resources


class MockPopen(Popen):
    def __init__(self, cmd: list, **kwargs):
        self.cmd = cmd
        self.args = kwargs
        self.env = kwargs.get("env")
        self.kernel_id = self.env.get("KERNEL_ID")
        self.resources = None
        self.pid = 42

    def wait(self, timeout=None):
        # This is called at cleanup and a good time to clear our resource cache
        assert self.resources
        self.resources.pop(self.kernel_id)
        return None

    def poll(self):
        # Ensure the resource still exits, else return non-None
        if self.resources and self.kernel_id in self.resources:
            return None
        return 2

    def mock_resources(self):
        """Sets up the initial resource (application, container) for discovery and state management"""
        if "GP_DOCKER_MODE" in self.env:  # This is docker, which one?
            resource = DockerResource(env=self.env)
            docker_resources[resource.kernel_id] = resource
            self.resources = docker_resources
        elif "KERNEL_POD_NAME" in self.env:  # This is k8s
            resource = K8sResource(env=self.env)
            k8s_resources[resource.kernel_id] = resource
            self.resources = k8s_resources
        elif "GP_IMPERSONATION_ENABLED" in self.env:  # This is yarn (but a little fragile)
            resource = YarnResource(env=self.env)
            yarn_resources[resource.kernel_id] = resource
            self.resources = yarn_resources
        else:
            err_msg = "Cant determine resource to mock!"
            raise AssertionError(err_msg)


def mock_launch_kernel(cmd: list, **kwargs) -> Popen:
    proc = MockPopen(cmd, **kwargs)
    proc.mock_resources()
    return proc



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Kernel managers that operate against a remote process."""
import asyncio
import errno
import getpass
import json
import os
import random
import re
import sys
import time
from abc import abstractmethod
from enum import Enum
from socket import AF_INET, SHUT_WR, SOCK_STREAM, socket, timeout
from typing import Any, Dict, List, Optional, Tuple

import pexpect
from jupyter_client import (
    KernelConnectionInfo,
    KernelProvisionerBase,
    launch_kernel,
    localinterfaces,
)
from overrides import overrides
from zmq.ssh import tunnel

from .config_mixin import (
    RemoteProvisionerConfigMixin,
    max_poll_attempts,
    poll_interval,
    socket_timeout,
    ssh_port,
)
from .response_manager import ResponseManager

# Pop certain env variables that don't need to be logged, e.g. remote_pwd
env_pop_list = ["GP_REMOTE_PWD", "LS_COLORS"]
default_kernel_launch_timeout = float(os.getenv("KERNEL_LAUNCH_TIMEOUT", "30"))

# Minimum port range size and max retries
min_port_range_size = int(os.getenv("GP_MIN_PORT_RANGE_SIZE", "1000"))
max_port_range_retries = int(os.getenv("GP_MAX_PORT_RANGE_RETRIES", "5"))

# Number of seconds in 100 years as the max keep-alive interval value.
max_keep_alive_interval_default = 100 * 365 * 24 * 60 * 60
max_keep_alive_interval = int(
    os.getenv("GP_TUNNEL_MAX_KEEP_ALIVE", max_keep_alive_interval_default)
)

tunneling_enabled = bool(os.getenv("GP_ENABLE_TUNNELING", "False").lower() == "true")

local_ip = localinterfaces.public_ips()[0]

random.seed()


class KernelChannel(Enum):
    """Enumeration used to better manage tunneling"""

    SHELL = "SHELL"
    IOPUB = "IOPUB"
    STDIN = "STDIN"
    HEARTBEAT = "HB"
    CONTROL = "CONTROL"
    COMMUNICATION = (
        "GP_COMM"  # Optional channel for remote launcher to issue interrupts - NOT a ZMQ channel
    )


def gp_launch_kernel(cmd: list, **kwargs):
    return launch_kernel(cmd, **kwargs)


class RemoteProvisionerBase(RemoteProvisionerConfigMixin, KernelProvisionerBase):
    """Base class for remote provisioners."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.start_time = None
        self.assigned_ip = None
        self.assigned_host = ""
        self.comm_ip = None
        self.comm_port = 0
        self.kernel_username = None
        self.tunneled_connect_info = None
        self.tunnel_processes = {}

        # Represents the local process (from popen) if applicable.  This will likely be non-None
        # for a short while until the script has launched the remote process, then will typically
        # go away.
        self.local_proc = None
        self.ip = None
        self.pid = 0
        self.pgid = 0

        self.response_manager = (
            ResponseManager.instance()
        )  # This will create the key pair and socket on first use
        self.response_address = self.response_manager.response_address
        self.public_key = self.response_manager.public_key
        self.lower_port, self.upper_port = self._validate_port_range()

    @property
    @abstractmethod
    @overrides
    def has_process(self) -> bool:
        pass

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        self.response_manager.register_event(self.kernel_id)

        cmd = self.kernel_spec.argv  # Build launch command, provide substitutions
        if self.response_address or self.port_range or self.kernel_id or self.public_key:
            ns = kwargs.copy()
            if self.response_address:
                ns["response_address"] = self.response_address
            if self.public_key:
                ns["public_key"] = self.public_key
            if self.port_range:
                ns["port_range"] = self.port_range
            if self.kernel_id:
                ns["kernel_id"] = self.kernel_id

            pat = re.compile(r"{([A-Za-z0-9_]+)}")

            def from_ns(match):
                """Get the key out of ns if it's there, otherwise no change."""
                return ns.get(match.group(1), match.group())

            cmd = [pat.sub(from_ns, arg) for arg in cmd]

        kwargs = await super().pre_launch(cmd=cmd, **kwargs)

        env = kwargs.get("env", {})
        self.kernel_username = env.get("KERNEL_USERNAME", getpass.getuser())  # Let env override
        env["KERNEL_USERNAME"] = self.kernel_username  # reset in env in case it's not there

        self._enforce_authorization(**kwargs)

        self.log.debug(f"RemoteProvisionerBase.pre_launch() env: {env}")
        return kwargs

    @overrides
    async def launch_kernel(self, cmd: List[str], **kwargs: Any) -> KernelConnectionInfo:
        launch_kwargs = RemoteProvisionerBase._scrub_kwargs(kwargs)
        self.local_proc = gp_launch_kernel(cmd, **launch_kwargs)
        self.pid = self.local_proc.pid
        self.ip = local_ip

        self.log_kernel_launch(cmd)

        await self.confirm_remote_startup()
        return self.connection_info

    @overrides
    async def post_launch(self, **kwargs: Any) -> None:
        pass

    @abstractmethod
    @overrides
    async def poll(self) -> Optional[int]:
        pass

    @overrides
    async def wait(self) -> Optional[int]:
        # If we have a local_proc, call its wait method.  This will cleanup any defunct processes when the kernel
        # is shutdown (when using waitAppCompletion = false).  Otherwise (if no local_proc) we'll use polling to
        # determine if a (remote or revived) process is still active.
        if self.local_proc:
            return self.local_proc.wait()

        poll_val = 0
        for _i in range(max_poll_attempts):
            poll_val = await self.poll()
            if poll_val is None:
                await asyncio.sleep(poll_interval)
            else:
                break
        else:
            self.log.warning(
                "Wait timeout of {} seconds exhausted. Continuing...".format(
                    max_poll_attempts * poll_interval
                )
            )
        return poll_val

    @overrides
    async def send_signal(self, signum: int) -> None:
        await self._send_signal_via_listener(signum)

    @abstractmethod
    @overrides
    async def kill(self, restart: bool = False) -> None:
        pass

    @abstractmethod
    @overrides
    async def terminate(self, restart: bool = False) -> None:
        pass

    @overrides
    async def cleanup(self, restart: bool = False) -> None:
        self.assigned_ip = None

        for kernel_channel, process in self.tunnel_processes.items():
            self.log.debug(f"cleanup: terminating {kernel_channel} tunnel process.")
            process.terminate()

        self.tunnel_processes.clear()

    @overrides
    async def shutdown_requested(self, restart: bool = False) -> None:
        await self.shutdown_listener(restart)

    @overrides
    async def get_provisioner_info(self) -> Dict[str, Any]:
        provisioner_info = await super().get_provisioner_info()
        provisioner_info.update(
            {
                "pid": self.pid,
                "pgid": self.pgid,
                "ip": self.ip,
                "assigned_ip": self.assigned_ip,
                "assigned_host": self.assigned_host,
                "comm_ip": self.comm_ip,
                "comm_port": self.comm_port,
                "tunneled_connect_info": self.tunneled_connect_info,
            }
        )
        return provisioner_info

    @overrides
    async def load_provisioner_info(self, provisioner_info: dict) -> None:
        await super().load_provisioner_info(provisioner_info)
        self.pid = provisioner_info.get("pid")
        self.pgid = provisioner_info.get("pgid")
        self.ip = provisioner_info.get("ip")
        self.assigned_ip = provisioner_info.get("assigned_ip")
        self.assigned_host = provisioner_info.get("assigned_host")
        self.comm_ip = provisioner_info.get("comm_ip")
        self.comm_port = provisioner_info.get("comm_port")
        self.tunneled_connect_info = provisioner_info.get("tunneled_connect_info")

    @overrides
    def get_shutdown_wait_time(self, recommended: Optional[float] = 5.0) -> float:
        return recommended

    @overrides
    def _finalize_env(self, env: Dict[str, str]) -> None:
        # add the applicable kernel_id and language to the env dict
        env["KERNEL_ID"] = self.kernel_id

        kernel_language = "unknown-kernel-language"
        if len(self.kernel_spec.language) > 0:
            kernel_language = self.kernel_spec.language.lower()
        # if already set in env: stanza, let that override.
        env["KERNEL_LANGUAGE"] = env.get("KERNEL_LANGUAGE", kernel_language)

        # Remove any potential sensitive (e.g., passwords) or annoying values (e.g., LG_COLORS)
        for k in env_pop_list:
            env.pop(k, None)

    @staticmethod
    def _scrub_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Remove any keyword arguments that Popen does not tolerate."""
        keywords_to_scrub: List[str] = ["extra_arguments", "kernel_id"]
        scrubbed_kwargs = kwargs.copy()
        for kw in keywords_to_scrub:
            scrubbed_kwargs.pop(kw, None)

        return scrubbed_kwargs

    @abstractmethod
    def log_kernel_launch(self, cmd: List[str]) -> None:
        """Logs the kernel launch from the respective remote provisioner"""
        pass

    async def handle_launch_timeout(self):
        """
        Checks to see if the kernel launch timeout has been exceeded while awaiting connection info.
        """
        await asyncio.sleep(poll_interval)
        time_interval = RemoteProvisionerBase.get_time_diff(self.start_time)

        if time_interval > self.launch_timeout:
            reason = f"Waited too long ({self.launch_timeout}s) to get connection file"
            timeout_message = f"KernelID: '{self.kernel_id}' launch timeout due to: {reason}"
            await self.kill()
            self.log_and_raise(TimeoutError(timeout_message))

    @abstractmethod
    async def confirm_remote_startup(self):
        """Confirms the remote process has started and returned necessary connection information."""
        pass

    def detect_launch_failure(self):
        """
        Helper method called from implementations of `confirm_remote_startup()` that checks if
        self.local_proc (a popen instance) has terminated prior to the confirmation of startup.
        This prevents users from having to wait for the kernel timeout duration to know if the
        launch fails.  It also helps distinguish local invocation issues from remote post-launch
        issues since the failure will be relatively immediate.

        Note that this method only applies to those process proxy implementations that launch
        from the local node.  Proxies like DistributedProcessProxy use rsh against a remote
        node, so there's not `local_proc` in play to interrogate.
        """

        # Check if the local proc has faulted (poll() will return non-None with a non-zero return
        # code in such cases).  If a fault was encountered, raise server error (500) with a message
        # indicating to check the EG log for more information.
        if self.local_proc:
            poll_result = self.local_proc.poll()
            if poll_result and poll_result > 0:
                self.local_proc.wait()  # FIXME
                error_message = (
                    f"Error occurred during launch of KernelID: {self.kernel_id}.  "
                    "Check Enterprise Gateway log for more information."
                )
                self.local_proc = None
                self.log_and_raise(RuntimeError(error_message))

    def log_and_raise(self, ex: Exception, chained: Optional[Exception] = None) -> None:
        """Helper method that logs the string-ized exception 'ex' and raises that exception.

        If a chained exception is provided that exception will be in the raised exception's from clause.

        :param ex : Exception
            The exception to log and raise
        :param chained : Exception (optional)
            The exception to use in the 'from' clause.
        """

        self.log.error(str(ex))
        if chained:
            raise ex from chained
        else:
            raise ex

    async def shutdown_listener(self, restart: bool) -> None:
        """
        Sends a shutdown request to the kernel launcher listener.
        """
        # If a comm port has been established, instruct the listener to shutdown so that proper
        # kernel termination can occur.  If not done, the listener keeps the launcher process
        # active, even after the kernel has terminated, leading to less than graceful terminations.

        if self.comm_port > 0:
            try:
                await self._send_listener_request({"shutdown": 1}, shutdown_socket=True)
                self.log.debug("Shutdown request sent to listener via gateway communication port.")
            except Exception as e:
                if not isinstance(e, OSError) or e.errno != errno.ECONNREFUSED:
                    self.log.warning(
                        f"An unexpected exception occurred sending listener shutdown "
                        f"to {self.comm_ip}:{self.comm_port} for "
                        f"KernelID '{self.kernel_id}': {e}"
                    )

            # Also terminate the tunnel process for the communication port - if in play.  Failure to terminate
            # this process results in the kernel (launcher) appearing to remain alive following the shutdown
            # request, which triggers the "forced kill" termination logic.

            comm_port_name = KernelChannel.COMMUNICATION.value
            comm_port_tunnel = self.tunnel_processes.get(comm_port_name, None)
            if comm_port_tunnel:
                self.log.debug(f"shutdown_listener: terminating {comm_port_name} tunnel process.")
                comm_port_tunnel.terminate()
                del self.tunnel_processes[comm_port_name]

    async def receive_connection_info(self) -> bool:
        """
        Monitors the response address for connection info sent by the remote kernel launcher.
        """
        # Polls the socket using accept.  When data is found, returns ready indicator and encrypted data.
        ready_to_connect = False
        try:
            connect_info = await self.response_manager.get_connection_info(self.kernel_id)
            self._setup_connection_info(connect_info)
            ready_to_connect = True
        except Exception as e:
            if (
                type(e) is timeout
                or type(e) is TimeoutError
                or type(e) is asyncio.exceptions.TimeoutError
            ):
                self.log.debug(
                    f"Waiting for KernelID '{self.kernel_id}' to send connection "
                    f"info from host '{self.assigned_host}' - retrying..."
                )
            else:
                error_message = (
                    f"Exception occurred waiting for connection file response for "
                    f"KernelId '{self.kernel_id}' on host '{self.assigned_host}': {e}"
                )
                await self.kill()
                self.log_and_raise(RuntimeError(error_message), chained=e)

        return ready_to_connect

    @staticmethod
    def get_current_time() -> int:
        """Return the current time (in milliseconds) from epoch.

        This method is intended for use in determining timeout values.
        """
        float_time = time.time()
        return int(float_time * 1000)  # Convert to ms and int

    @staticmethod
    def get_time_diff(start_time_ms: int) -> float:
        """Return the difference (in seconds) between the given start_time and the current time"""
        end_time_ms = RemoteProvisionerBase.get_current_time()
        time_diff = float((end_time_ms - start_time_ms) / 1000)
        return time_diff

    @staticmethod
    def ip_is_local(ip):
        """Returns True if `ip` is considered local to this server, False otherwise."""
        return localinterfaces.is_public_ip(ip) or localinterfaces.is_local_ip(ip)

    async def _send_signal_via_listener(self, signum: int) -> bool:
        """Sends signal 'signum' to kernel process via listener.

        :returns: True if request delivered, false otherwise.
        """
        # If the launcher returned a comm_port value, then use that to send the signal,
        # else, defer to the superclass - which will use a remote shell to issue kill.
        # Note that if the target process is running as a different user than the REMOTE_USER,
        # using anything other than the socket-based signal (via signal_addr) will not work.
        if self.comm_port > 0:
            try:
                await self._send_listener_request({"signum": signum})
                if signum > 0:  # Polling (signum == 0) is too frequent
                    self.log.debug(f"Signal ({signum}) sent via gateway communication port.")
                return True
            except Exception as e:
                if (
                    isinstance(e, OSError) and e.errno == errno.ECONNREFUSED
                ):  # Return False since there's no process.
                    if (
                        signum > 0
                    ):  # Since poll's purpose is to check status, consider this as terminated.
                        self.log.debug(
                            "ERROR: ECONNREFUSED, no process listening, cannot send signal."
                        )
                else:
                    self.log.warning(
                        f"An unexpected exception occurred sending signal ({signum}) "
                        f"via listener for KernelID '{self.kernel_id}': {e}"
                    )
        return False

    def _enforce_authorization(self, **kwargs):
        """Applies any authorization configuration using the kernel user.

        Authorization is performed by comparing the value of KERNEL_USERNAME with each value in the set of
        unauthorized users.  If any (case-sensitive) matches are found, HTTP error 403 (Forbidden) will be raised
        - preventing the launch of the kernel.  If the authorized_users set is non-empty, it is then checked to
        ensure the value of KERNEL_USERNAME is present in that list.  If not found, HTTP error 403 will be raised.

        """

        # Now perform authorization checks
        if self.kernel_username in self.unauthorized_users:
            self._raise_authorization_error("not authorized")

        # If authorized users are non-empty, ensure user is in that set.
        if self.authorized_users.__len__() > 0:
            if self.kernel_username not in self.authorized_users:
                self._raise_authorization_error("not in the set of users authorized")

    def _raise_authorization_error(self, differentiator_clause):
        """Raises a 403 status code after building the appropriate message."""
        kernel_name = self.kernel_spec.display_name
        kernel_clause = f" '{kernel_name}'." if kernel_name is not None else "s."
        error_message = (
            f"User '{self.kernel_username}' is {differentiator_clause} to start kernel{kernel_clause} "
            "Ensure KERNEL_USERNAME is set to an appropriate value and retry the request."
        )
        self.log_and_raise(PermissionError(error_message))

    def _validate_port_range(self) -> Tuple[int, int]:
        """Validates the port range configuration option to ensure appropriate values."""

        lower_port = upper_port = 0
        port_range = self.port_range
        try:
            port_ranges = port_range.split("..")

            lower_port = int(port_ranges[0])
            upper_port = int(port_ranges[1])

            port_range_size = upper_port - lower_port
            if port_range_size != 0:
                if port_range_size < min_port_range_size:
                    self.log_and_raise(
                        ValueError(
                            f"Port range validation failed for range: '{port_range}'.  "
                            f"Range size must be at least {min_port_range_size} as "
                            f"specified by env GP_MIN_PORT_RANGE_SIZE"
                        )
                    )

                # According to RFC 793, port is a 16-bit unsigned int. Which means the port
                # numbers must be in the range (0, 65535). However, within that range,
                # ports 0 - 1023 are called "well-known ports" and are typically reserved for
                # specific purposes. For example, 0 is reserved for random port assignment,
                # 80 is used for HTTP, 443 for TLS/SSL, 25 for SMTP, etc. But, there is
                # flexibility as one can choose any port with the aforementioned protocols.
                # Ports 1024 - 49151 are called "user or registered ports" that are bound to
                # services running on the server listening to client connections. And, ports
                # 49152 - 65535 are called "dynamic or ephemeral ports". A TCP connection
                # has two endpoints. Each endpoint consists of an IP address and a port number.
                # And, each connection is made up of a 4-tuple consisting of -- client-IP,
                # client-port, server-IP, and server-port. A service runs on a server with a
                # specific IP and is bound to a specific "user or registered port" that is
                # advertised for clients to connect. So, when a client connects to a service
                # running on a server, three out of 4-tuple - client-IP, client-port, server-IP -
                # are already known. To be able to serve multiple clients concurrently, the
                # server's IP stack assigns an ephemeral port for the connection to complete
                # the 4-tuple.
                #
                # In case of JEG, we will accept ports in the range 1024 - 65535 as these days
                # admins use dedicated hosts for individual services.
                def validate_port(port: int) -> None:
                    if port < 1024 or port > 65535:
                        self.log_and_raise(
                            ValueError(
                                f"Invalid port range '{port_range}' specified. "
                                "Range for valid port numbers is (1024, 65535)."
                            )
                        )

                validate_port(lower_port)
                validate_port(upper_port)
        except IndexError as ie:
            self.log_and_raise(
                RuntimeError(f"Port range validation failed for range: '{port_range}'."),
                chained=ie,
            )

        return lower_port, upper_port

    def _setup_connection_info(self, connect_info: dict) -> None:
        """
        Take connection info (returned from launcher or loaded from session persistence) and properly
        configure port variables for the 5 kernel and (possibly) the launcher communication port.  If
        tunneling is enabled, these ports will be tunneled with the original port information recorded.
        """

        self.log.debug(
            f"Host assigned to the kernel is: '{self.assigned_host}' '{self.assigned_ip}'"
        )

        connect_info[
            "ip"
        ] = self.assigned_ip  # Set connection to IP address of system where the kernel was launched

        if tunneling_enabled is True:
            # Capture the current(tunneled) connect_info relative to the IP and ports (including the
            # communication port - if present).
            self.tunneled_connect_info = dict(connect_info)

            # Open tunnels to the 5 ZMQ kernel ports
            tunnel_ports = self._tunnel_to_kernel(connect_info, self.assigned_ip)
            self.log.debug(f"Local ports used to create SSH tunnels: '{tunnel_ports}'")

            # Replace the remote connection ports with the local ports used to create SSH tunnels.
            connect_info["ip"] = "127.0.0.1"
            connect_info["shell_port"] = tunnel_ports[0]
            connect_info["iopub_port"] = tunnel_ports[1]
            connect_info["stdin_port"] = tunnel_ports[2]
            connect_info["hb_port"] = tunnel_ports[3]
            connect_info["control_port"] = tunnel_ports[4]

            # If a communication port was provided, tunnel it
            if "comm_port" in connect_info:
                self.comm_ip = connect_info["ip"]
                tunneled_comm_port = int(connect_info["comm_port"])
                self.comm_port = self._tunnel_to_port(
                    KernelChannel.COMMUNICATION,
                    self.assigned_ip,
                    tunneled_comm_port,
                    self.assigned_ip,
                )
                connect_info["comm_port"] = self.comm_port
                self.log.debug(
                    f"Established communication to: {self.assigned_ip}:{tunneled_comm_port} "
                    f"for KernelID '{self.kernel_id}' via tunneled port 127.0.0.1:{self.comm_port}"
                )

        else:  # tunneling not enabled, still check for and record communication port
            if "comm_port" in connect_info:
                self.comm_ip = connect_info["ip"]
                self.comm_port = int(connect_info["comm_port"])
                self.log.debug(
                    f"Established communication to: {self.assigned_ip}:{self.comm_port} "
                    f"for KernelID '{self.kernel_id}'"
                )

        # If no communication port was provided, record that fact as well since this is useful to know
        if "comm_port" not in connect_info:
            self.log.debug(
                f"Communication port has NOT been established for KernelID '{self.kernel_id}' (optional)."
            )

        self._update_connection(connect_info)

    def _update_connection(self, connect_info: dict) -> None:
        """
        Updates the connection info with that received from launcher.  Also pulls the PID and PGID
        info, if present, in case we need to use it for lifecycle management.
        Note: Do NOT update connect_info with IP and other such artifacts in this method/function.
        """

        if not connect_info:
            error_message = (
                f"Unexpected runtime encountered for Kernel ID '{self.kernel_id}' - "
                f"connection information is null!"
            )
            self.log_and_raise(RuntimeError(error_message))

        # Load new connection information into memory. No need to write back out to a file or track loopback, etc.
        # The launcher may also be sending back process info, so check and extract
        self._extract_pid_info(connect_info)
        self.log.debug(
            f"Received connection info for KernelID '{self.kernel_id}' "
            f"from host '{self.assigned_host}': {connect_info}..."
        )

        self.connection_info.update(connect_info)

    def _extract_pid_info(self, connect_info: dict) -> None:
        """
        Extracts any PID, PGID info from the payload received on the response socket.
        """
        pid = connect_info.pop("pid", None)
        if pid:
            try:
                self.pid = int(pid)
            except ValueError:
                self.log.warning(
                    f"pid returned from kernel launcher is not an integer: {pid} - ignoring."
                )
                pid = None
        pgid = connect_info.pop("pgid", None)
        if pgid:
            try:
                self.pgid = int(pgid)
            except ValueError:
                self.log.warning(
                    f"pgid returned from kernel launcher is not an integer: {pgid} - ignoring."
                )
                pgid = None
        if (
            pid or pgid
        ):  # if either process ids were updated, update the ip as well and don't use local_proc
            self.ip = self.assigned_ip
            if not RemoteProvisionerBase.ip_is_local(
                self.ip
            ):  # only unset local_proc if we're remote
                # FIXME - should we wait prior to unset?
                self.local_proc = None

    async def _send_listener_request(
        self, request: dict, shutdown_socket: Optional[bool] = False
    ) -> None:
        """
        Sends the request dictionary to the kernel listener via the comm port.  Caller is responsible for
        handling any exceptions.
        """
        if self.comm_port > 0:
            sock = socket(AF_INET, SOCK_STREAM)
            try:
                sock.settimeout(socket_timeout)
                await asyncio.get_event_loop().sock_connect(sock, (self.comm_ip, self.comm_port))
                sock.send(json.dumps(request).encode(encoding="utf-8"))
            finally:
                if shutdown_socket:
                    try:
                        sock.shutdown(SHUT_WR)
                    except Exception as e2:
                        issue_warning: bool = True
                        if isinstance(e2, OSError) and e2.errno == errno.ENOTCONN:
                            # Listener is not connected.  This is probably a follow-on to ECONNREFUSED on connect
                            if (
                                "shutdown" in request
                            ):  # If this is a shutdown request, dampen the urgency
                                issue_warning = False
                                self.log.debug(
                                    "OSError(ENOTCONN) raised on socket shutdown, listener likely "
                                    "not connected due to listener's expected termination."
                                )
                        if issue_warning:
                            self.log.warning(
                                f"Exception occurred attempting to shutdown communication "
                                f"socket to {self.comm_ip}:{self.comm_port} for KernelID "
                                f"'{self.kernel_id}' (ignored): {str(e2)}"
                            )
                sock.close()
        else:
            self.log.debug(
                "Invalid comm port, not sending request '{}' to comm_port '{}'.",
                request,
                self.comm_port,
            )

    def _tunnel_to_kernel(
        self,
        connection_info: dict,
        server: str,
        port: int = ssh_port,
        key: Optional[str] = None,
    ):
        """
        Tunnel connections to a kernel over SSH

        This will open five SSH tunnels from localhost on this machine to the
        ports associated with the kernel.
        See jupyter_client/connect.py for original implementation.
        """
        cf = connection_info

        lports = self._select_ports(5)

        rports = (
            cf["shell_port"],
            cf["iopub_port"],
            cf["stdin_port"],
            cf["hb_port"],
            cf["control_port"],
        )

        channels = (
            KernelChannel.SHELL,
            KernelChannel.IOPUB,
            KernelChannel.STDIN,
            KernelChannel.HEARTBEAT,
            KernelChannel.CONTROL,
        )

        remote_ip = cf["ip"]

        if not tunnel.try_passwordless_ssh(server + ":" + str(port), key):
            self.log_and_raise(
                PermissionError(
                    "Must use password-less scheme by setting up the "
                    "SSH public key on the cluster nodes"
                )
            )

        for lp, rp, kc in zip(lports, rports, channels):
            self._create_ssh_tunnel(kc, lp, rp, remote_ip, server, port, key)

        return tuple(lports)

    def _tunnel_to_port(
        self,
        kernel_channel: KernelChannel,
        remote_ip: str,
        remote_port: int,
        server: str,
        port: int = ssh_port,
        key: Optional[str] = None,
    ):
        """
        Analogous to _tunnel_to_kernel, but deals with a single port.  This will typically be called for
        any one-off ports that require tunnelling. Note - this method assumes that passwordless ssh is
        in use and has been previously validated.
        """
        local_port = self._select_ports(1)[0]
        self._create_ssh_tunnel(
            kernel_channel, local_port, remote_port, remote_ip, server, port, key
        )
        return local_port

    def _create_ssh_tunnel(
        self,
        kernel_channel: KernelChannel,
        local_port: int,
        remote_port: int,
        remote_ip: str,
        server: str,
        port: int,
        key: Optional[str] = None,
    ):
        """
        Creates an SSH tunnel between the local and remote port/server for the given kernel channel.
        """
        channel_name = kernel_channel.value
        self.log.debug(
            f"Creating SSH tunnel for '{channel_name}': 127.0.0.1:'{local_port}' "
            f"to '{remote_ip}':'{remote_port}'"
        )
        try:
            process = RemoteProvisionerBase._spawn_ssh_tunnel(
                local_port, remote_port, remote_ip, server, port, key
            )
            self.tunnel_processes[channel_name] = process
        except Exception as e:
            self.log_and_raise(
                RuntimeError(
                    f"Could not open SSH tunnel for port {channel_name}. Exception: '{e}'"
                ),
                chained=e,
            )

    @staticmethod
    def _spawn_ssh_tunnel(
        local_port: int,
        remote_port: int,
        remote_ip: str,
        server: str,
        port: int,
        key: Optional[str] = None,
    ):
        """
        This method spawns a child process to create an SSH tunnel and returns the spawned process.
        ZMQ's implementation returns a pid on UNIX based platforms and a process handle/reference on
        Win32. By consistently returning a process handle/reference on both UNIX and Win32 platforms,
        this method enables the caller to deal with the same currency regardless of the platform. For
        example, on both UNIX and Win32 platforms, the developer will have the option to stash the
        child process reference and manage it's lifecycle consistently.

        On UNIX based platforms, ZMQ's implementation is more generic to be able to handle various
        use-cases. ZMQ's implementation also requests the spawned process to go to background using
        '-f' command-line option. As a result, the spawned process becomes an orphan and any references
        to the process obtained using it's pid become stale. On the other hand, this implementation is
        specifically for password-less SSH login WITHOUT the '-f' command-line option thereby allowing
        the spawned process to be owned by the parent process. This allows the parent process to control
        the lifecycle of it's child processes and do appropriate cleanup during termination.
        """
        if sys.platform == "win32":
            ssh_server = server + ":" + str(port)
            return tunnel.paramiko_tunnel(local_port, remote_port, ssh_server, remote_ip, key)
        else:
            ssh = "ssh -p %s -o ServerAliveInterval=%i" % (
                port,
                max_keep_alive_interval,
            )
            cmd = "%s -S none -L 127.0.0.1:%i:%s:%i %s" % (
                ssh,
                local_port,
                remote_ip,
                remote_port,
                server,
            )
            return pexpect.spawn(cmd, env=os.environ.copy().pop("SSH_ASKPASS", None))

    def _select_ports(self, count: int) -> List[int]:
        """
        Selects and returns n random ports that adhere to the configured port range, if applicable.

        Parameters
        ----------
        count : int
            The number of ports to return

        Returns
        -------
        List - ports available and adhering to the configured port range
        """
        ports: List[int] = []
        sockets: List[socket] = []
        for _i in range(count):
            sock = self._select_socket()
            ports.append(sock.getsockname()[1])
            sockets.append(sock)
        for sock in sockets:
            sock.close()
        return ports

    def _select_socket(self, ip: str = "") -> socket:
        """
        Creates and returns a socket whose port adheres to the configured port range, if applicable.

        Parameters
        ----------
        ip : str
            Optional ip address to which the port is bound

        Returns
        -------
        socket - Bound socket that is available and adheres to configured port range
        """
        sock = socket(AF_INET, SOCK_STREAM)
        found_port = False
        retries = 0
        while not found_port:
            try:
                sock.bind((ip, self._get_candidate_port()))
                found_port = True
            except Exception:
                retries = retries + 1
                if retries > max_port_range_retries:
                    self.log_and_raise(
                        RuntimeError(
                            f"Failed to locate port within range {self.port_range} "
                            f"after {max_port_range_retries} retries!"
                        )
                    )
        return sock

    def _get_candidate_port(self):
        """Randomly selects a port number within the configured range.

        If no range is configured, the 0 port is used - allowing the server to choose from the full range.
        """
        range_size = self.upper_port - self.lower_port
        if range_size == 0:
            return 0
        return random.randint(self.lower_port, self.upper_port)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
###
# NOTE: This is a placeholder file to satisfy references in the corresponding launch_ipykernel.py
# file.  It will be replaced with the file in ../../shared/server_listener.py during kernel-launcher
# assembly.  The file is also used by the launch_IRkernel.R script, but it does not require a
# placeholder to satisfy references due to the language differences.
#
from typing import Optional


def setup_server_listener(
    conn_filename: str,
    parent_pid: int,
    lower_port: int,
    upper_port: int,
    response_addr: str,
    kernel_id: str,
    public_key: str,
    cluster_type: Optional[str] = None,
    as_thread: Optional[bool] = True,
):
    err_msg = "kernel-launcher assembly is required!"
    raise NotImplementedError(err_msg)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys
from socket import socket

import pytest
import yarn_api_client

os.environ["PYTEST_CURRENT_TEST"] = "1"
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"  # Avoid deprecation warning and use the platform dirs now

# See compatibility note on `group` keyword in https://docs.python.org/3/library/importlib.metadata.html#entry-points
if sys.version_info < (3, 10):  # pragma: no cover
    from importlib_metadata import EntryPoint, entry_points
else:  # pragma: no cover
    from importlib.metadata import EntryPoint, entry_points
from docker.client import DockerClient
from jupyter_client.kernelspec import KernelSpec
from jupyter_client.provisioning.factory import KernelProvisionerFactory
from mocks.docker_client import mock_docker_from_env
from mocks.k8s_client import MockK8sClient
from mocks.popen import mock_launch_kernel
from mocks.response_manager import mock_get_connection_info, mock_register_event, mock_socket_listen
from mocks.yarn_client import MockResourceManager

import gateway_provisioners
from gateway_provisioners.k8s import client  # noqa: F401
from gateway_provisioners.remote_provisioner import RemoteProvisionerBase
from gateway_provisioners.response_manager import ResponseManager


@pytest.fixture
def response_manager(monkeypatch):
    """Setup the Kernel Provisioner Factory, mocking the entrypoint fetch calls."""
    monkeypatch.setattr(ResponseManager, "register_event", mock_register_event)
    monkeypatch.setattr(socket, "listen", mock_socket_listen)
    monkeypatch.setattr(ResponseManager, "get_connection_info", mock_get_connection_info)
    rm = ResponseManager.instance()
    yield rm
    ResponseManager.clear_instance()


@pytest.fixture
def init_api_mocks(monkeypatch):
    monkeypatch.setattr(DockerClient, "from_env", mock_docker_from_env)
    monkeypatch.setattr(yarn_api_client.resource_manager, "ResourceManager", MockResourceManager)
    monkeypatch.setattr(
        gateway_provisioners.remote_provisioner, "gp_launch_kernel", mock_launch_kernel
    )
    monkeypatch.setattr(gateway_provisioners.k8s, "client", MockK8sClient)


@pytest.fixture
def kernelspec():
    def _kernelspec(name: str) -> KernelSpec:
        kspec = KernelSpec()
        kspec.argv = [
            "--public-key:{public_key}",
            "--response-address:{response_address}",
            "--port-range:{port_range}",
            "--kernel-id:{kernel_id}",
        ]
        kspec.display_name = f"{name}_python"
        kspec.language = "python"
        kspec.env = {}
        kspec.metadata = {}
        return kspec

    return _kernelspec


@pytest.fixture
def get_provisioner(kernelspec):
    def _get_provisioner(name: str, kernel_id: str) -> RemoteProvisionerBase:
        provisioner_config = {}
        provisioner_name = name + "-provisioner"
        eps = entry_points(group=KernelProvisionerFactory.GROUP_NAME, name=provisioner_name)
        assert eps, f"No entry_point was returned for provisioner '{provisioner_name}'!"
        ep: EntryPoint = eps[provisioner_name]
        provisioner_class = ep.load()
        provisioner: RemoteProvisionerBase = provisioner_class(
            kernel_id=kernel_id, kernel_spec=kernelspec(name), parent=None, **provisioner_config
        )
        return provisioner

    return _get_provisioner



"""Code related to managing kernels running based on k8s custom resource."""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import logging
import re
from contextlib import suppress
from typing import Any, Dict, List, Optional, Set

from overrides import overrides

try:
    from kubernetes import client
except ImportError:
    logging.warning(
        "At least one of the extra packages 'kubernetes' or 'jinja2' are not installed in "
        "this environment and are required.  Ensure that gateway_provisioners is installed "
        "by specifying the extra 'k8s' (e.g., pip install 'gateway_provisioners[k8s]')."
    )
    raise

from .k8s import KubernetesProvisioner


class CustomResourceProvisioner(KubernetesProvisioner):
    """A custom resource provisioner."""

    # Identifies the kind of object being managed by this provisioner.
    # For these values we will prefer the values found in the 'kind' field
    # of the object's metadata.  This attribute is strictly used to provide
    # context to log messages.
    object_kind = "CustomResourceDefinition"

    def __init__(self, **kwargs):
        """Initialize the provisioner."""
        super().__init__(**kwargs)
        self.group = self.version = self.plural = None
        self.kernel_resource_name = None

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        """Launch the process for a kernel."""
        kwargs = await super().pre_launch(**kwargs)
        self.kernel_resource_name = self._determine_kernel_pod_name(**kwargs)
        kwargs["env"]["KERNEL_RESOURCE_NAME"] = self.kernel_resource_name
        kwargs["env"]["KERNEL_CRD_GROUP"] = self.group
        kwargs["env"]["KERNEL_CRD_VERSION"] = self.version
        kwargs["env"]["KERNEL_CRD_PLURAL"] = self.plural
        return kwargs

    @overrides
    def get_container_status(self, iteration: Optional[str]) -> str:
        """Determines submitted CRD application status

        Submitting a new kernel application CRD will take a while to
        reach the running state and the submission can also fail due
        to malformation or other issues which will prevent the application
        pod to reach the desired running state.

        This function checks the CRD submission state and in case of
        success it then delegates to parent to check if the application
        pod is running.

        Returns
        -------
        Empty string if the container cannot be found otherwise.
        The pod application status in case of success on Spark Operator side
        Or the retrieved spark operator submission status in other cases (e.g. Failed)
        """

        application_state = ""

        with suppress(Exception):
            custom_resource = client.CustomObjectsApi().get_namespaced_custom_object(
                self.group,
                self.version,
                self.kernel_namespace,
                self.plural,
                self.kernel_resource_name,
            )

            if custom_resource:
                application_state = custom_resource["status"]["applicationState"]["state"].lower()

                if application_state in self.get_error_states():
                    exception_text = CustomResourceProvisioner._get_exception_text(
                        custom_resource["status"]["applicationState"]["errorMessage"]
                    )
                    error_message = (
                        f"CRD submission for kernel {self.kernel_id} failed: {exception_text}"
                    )
                    self.log.debug(error_message)
                elif application_state == "running" and not self.assigned_host:
                    application_state = super().get_container_status(iteration)

        # only log if iteration is not None (otherwise poll() is too noisy)
        # check for running state to avoid double logging with superclass
        if iteration and application_state != "running":
            self.log.debug(
                f"{iteration}: Waiting from CRD status from resource manager {self.object_kind.lower()} in "
                f"namespace '{self.kernel_namespace}'. Name: '{self.kernel_resource_name}', "
                f"Status: '{application_state}', KernelID: '{self.kernel_id}'"
            )

        return application_state

    @overrides
    def delete_managed_object(self, termination_stati: List[str]) -> bool:
        """Deletes the object managed by this provisioner

        A return value of True indicates the object is considered deleted,
        otherwise a False or None value is returned.

        Note: the caller is responsible for handling exceptions.
        """
        delete_status = client.CustomObjectsApi().delete_namespaced_custom_object(
            self.group,
            self.version,
            self.kernel_namespace,
            self.plural,
            self.kernel_resource_name,
            grace_period_seconds=0,
            propagation_policy="Background",
        )

        result = delete_status and delete_status.get("status", None) in termination_stati

        return result

    @overrides
    def get_initial_states(self) -> Set[str]:
        """Return list of states in lowercase indicating container is starting (includes running)."""
        return {"submitted", "pending", "running"}

    @staticmethod
    def _get_exception_text(error_message) -> str:
        match = re.search(r"Exception\s*:\s*(.*)", error_message, re.MULTILINE)

        if match:
            error_message = match.group(1)

        return error_message



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from time import time_ns
from typing import Optional


class YarnResource:
    """Track the current state of the resource"""

    initial_states = {"NEW", "SUBMITTED", "ACCEPTED", "RUNNING"}
    final_states = {"FINISHED", "KILLED", "FAILED"}

    def __init__(self, env: dict):
        self.env = env
        self.kernel_id: str = env.get("KERNEL_ID")
        self.kernel_username: str = env.get("KERNEL_USERNAME")
        self.name = self.kernel_id
        self.id = f"application_{str(time_ns())}_0001"
        self.status: str = "NEW"
        self.query_counter: int = 1
        self.terminate_counter: int = 0


yarn_resources: dict = {}


class MockResponse:
    def __init__(
        self, apps: Optional[dict] = None, app: Optional[dict] = None, status: Optional[str] = None
    ):
        self.data = {}
        if apps:
            self.data["apps"] = apps
        elif app:
            self.data["app"] = app
        elif status:
            self.data["status"] = status


class MockResourceManager:
    CLUSTER_CONTAINER_MEMORY = 1024 * 1024 * 1024  # 1GB

    def __init__(self, **kwargs):
        self.endpoints = kwargs.get("service_endpoints")

    def get_active_endpoint(self):
        assert len(self.endpoints) > 0
        return self.endpoints[0]

    def cluster_applications(
        self,
        state=None,
        states=None,
        final_status=None,
        user=None,
        queue=None,
        limit=None,
        started_time_begin=None,
        started_time_end=None,
        finished_time_begin=None,
        finished_time_end=None,
        application_types=None,
        application_tags=None,
        name=None,
        de_selects=None,
    ):
        """This method is used to determine when the application ID has been created"""
        apps = {"app": []}
        app_list = apps.get("app")
        for kid, resource in yarn_resources.items():
            # convert each resource into an app list entry
            id = ""
            if resource.query_counter >= 3:
                id = resource.id
                resource.status = "RUNNING"
            resource.query_counter += 1
            app_entry: dict = {"name": kid, "id": id, "state": resource.status}
            app_list.append(app_entry)
        response = MockResponse(apps=apps)
        return response

    def cluster_application(self, application_id):
        response = MockResponse()
        resource = MockResourceManager._locate_resource(application_id)
        if resource:
            app_entry: dict = {
                "name": resource.kernel_id,
                "id": resource.id,
                "state": resource.status,
                "amHostHttpAddress": "localhost:8042",
            }
            response.data["app"] = app_entry

        return response

    def cluster_application_state(self, application_id):
        response = MockResponse()
        resource = MockResourceManager._locate_resource(application_id)

        if resource:
            if resource.terminate_counter:  # Let this cycle a bit
                if resource.terminate_counter > 3:
                    resource.status = "FINISHED"
                resource.terminate_counter += 1
            response.data["state"] = resource.status

        return response

    def cluster_application_kill(self, application_id):
        response = MockResponse()
        resource = MockResourceManager._locate_resource(application_id)
        if resource:
            response.data["state"] = resource.status
            resource.terminate_counter = 1

    def cluster_node_container_memory(self):
        return MockResourceManager.CLUSTER_CONTAINER_MEMORY

    def cluster_scheduler_queue(self, yarn_queue_name):
        # TODO - add impl when adding queue testing
        pass

    def cluster_queue_partition(self, yarn_queue, node_label):
        # TODO - add impl when adding queue testing
        pass

    def cluster_scheduler_queue_availability(self, partition, partition_availability_threshold):
        # TODO - add impl when adding queue testing
        pass

    @staticmethod
    def _locate_resource(app_id: str) -> Optional[YarnResource]:
        for resource in yarn_resources.values():
            if resource.id == app_id:
                return resource
        return None



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Code related to managing kernels running in docker-based containers."""
import logging
import os
from typing import Any, Dict, Optional, Set

from overrides import overrides

try:
    from docker.client import DockerClient
    from docker.errors import NotFound
    from docker.models.containers import Container
    from docker.models.services import Service
except ImportError:
    logging.warning(
        "The extra package 'docker' is not installed in this environment and is required.  "
        "Ensure that gateway_provisioners is installed by specifying the extra 'docker' "
        "(e.g., pip install 'gateway_provisioners[docker]')."
    )
    raise

from .container import ContainerProvisionerBase

# Debug logging level of docker produces too much noise - raise to info by default.
logging.getLogger("urllib3.connectionpool").setLevel(
    os.environ.get("GP_DOCKER_LOG_LEVEL", logging.WARNING)
)

docker_network = os.environ.get("GP_DOCKER_NETWORK", "bridge")


class DockerSwarmProvisioner(ContainerProvisionerBase):
    """Kernel provisioner for kernels in Docker Swarm."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.client = DockerClient.from_env()

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        kwargs = await super().pre_launch(**kwargs)

        # Convey the network to the docker launch script
        kwargs["env"]["GP_DOCKER_NETWORK"] = docker_network
        kwargs["env"]["GP_DOCKER_MODE"] = "swarm"
        return kwargs

    @overrides
    def get_initial_states(self) -> Set[str]:
        return {"preparing", "starting", "running"}

    @overrides
    def get_error_states(self) -> Set[str]:
        return {"failed", "rejected", "complete", "shutdown", "orphaned", "remove"}

    @overrides
    def get_container_status(self, iteration: Optional[str]) -> str:
        # Locates the kernel container using the kernel_id filter.  If the status indicates an initial state we
        # should be able to get at the NetworksAttachments and determine the associated container's IP address.
        task_state = ""
        task_id = None
        task = self._get_task()
        if task:
            task_status = task["Status"]
            task_id = task["ID"]
            if task_status:
                task_state = task_status["State"].lower()
                if (
                    self.assigned_host == "" and task_state == "running"
                ):  # in self.get_initial_states():
                    # get the NetworkAttachments and pick out the first of the Network and first
                    networks_attachments = task["NetworksAttachments"]
                    if len(networks_attachments) > 0:
                        address = networks_attachments[0]["Addresses"][0]
                        ip = address.split("/")[0]
                        self.assigned_ip = ip
                        self.assigned_host = self.container_name

        if iteration:  # only log if iteration is not None (otherwise poll() is too noisy)
            self.log.debug(
                f"{iteration}: Waiting to connect to docker container. "
                f"Name: '{self.container_name}', Status: '{task_state}', "
                f"IPAddress: '{self.assigned_ip}', KernelID: '{self.kernel_id}', "
                f"TaskID: '{task_id}'"
            )
        return task_state

    @overrides
    def terminate_container_resources(self, restart: bool = False) -> Optional[bool]:
        # Remove the docker service.

        result = True  # We'll be optimistic
        service = self._get_service()
        if service:
            try:
                service.remove()  # Service still exists, attempt removal
            except Exception as err:
                self.log.debug(
                    f"{self.__class__.__name__} Termination of service: "
                    f"{service.name} raised exception: {err}"
                )
                if isinstance(err, NotFound):
                    pass  # okay if its not found
                else:
                    result = False
                    self.log.warning(f"Error occurred removing service: {err}")
        if result:
            self.log.debug(
                f"{self.__class__.__name__}.terminate_container_resources, "
                f"service {self.container_name}, "
                f"kernel ID: {self.kernel_id} has been terminated."
            )
            self.container_name = None
            result = None  # maintain jupyter contract
        else:
            self.log.warning(
                f"{self.__class__.__name__}.terminate_container_resources, "
                f"container {self.container_name}, "
                f"kernel ID: {self.kernel_id} has not been terminated."
            )
        return result

    def _get_service(self) -> Service:
        """Fetches the service object corresponding to the kernel with a matching label."""
        service = None
        services = self.client.services.list(filters={"label": "kernel_id=" + self.kernel_id})
        num_services = len(services)
        if num_services != 1:
            if num_services > 1:
                err_msg = (
                    f"{self.__class__.__name__}: Found more than one service "
                    f"({num_services}) for kernel_id '{self.kernel_id}'!"
                )
                raise RuntimeError(err_msg)
        else:
            service = services[0]
            self.container_name = service.name
        return service

    def _get_task(self) -> dict:
        """
        Fetches the task object corresponding to the service associated with the kernel.  We only ask for the
        current task with desired-state == running.  This eliminates failed states.
        """
        task = None
        service = self._get_service()
        if service:
            tasks = service.tasks(filters={"desired-state": "running"})
            num_tasks = len(tasks)
            if num_tasks != 1:
                if num_tasks > 1:
                    err_msg = (
                        f"{self.__class__.__name__}: Found more than one task ({num_tasks}) "
                        f"for service '{service.name}', kernel_id '{self.kernel_id}'!"
                    )
                    raise RuntimeError(err_msg)
            else:
                task = tasks[0]
        return task


class DockerProvisioner(ContainerProvisionerBase):
    """Kernel provisioner for kernels in Docker (non-Swarm)."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.client = DockerClient.from_env()

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        kwargs = await super().pre_launch(**kwargs)

        # Convey the network to the docker launch script
        kwargs["env"]["GP_DOCKER_NETWORK"] = docker_network
        kwargs["env"]["GP_DOCKER_MODE"] = "docker"
        return kwargs

    @overrides
    def get_initial_states(self) -> Set[str]:
        return {"created", "running"}

    @overrides
    def get_error_states(self) -> Set[str]:
        return {"restarting", "paused", "exited", "dead"}

    @overrides
    def get_container_status(self, iteration: Optional[str]) -> str:
        # Locates the kernel container using the kernel_id filter.  If the phase indicates Running, the pod's IP
        # is used for the assigned_ip.  Only used when docker mode == regular (non swarm)
        container_status = ""

        container = self._get_container()
        if container:
            self.container_name = container.name
            if container.status:
                container_status = container.status.lower()
                if container_status == "running" and self.assigned_host == "":
                    # Container is running, capture IP

                    # we'll use this as a fallback in case we don't find our network
                    self.assigned_ip = container.attrs.get("NetworkSettings").get("IPAddress")
                    networks = container.attrs.get("NetworkSettings").get("Networks")
                    if len(networks) > 0:
                        self.assigned_ip = networks.get(docker_network).get("IPAddress")
                        self.log.debug(
                            f"Using assigned_ip {self.assigned_ip} from docker network '{docker_network}'."
                        )
                    else:
                        self.log.warning(
                            f"Docker network '{docker_network}' could not be located in "
                            f"container attributes - using assigned_ip '{self.assigned_ip}'."
                        )

                    self.assigned_host = self.container_name

        if iteration:  # only log if iteration is not None (otherwise poll() is too noisy)
            self.log.debug(
                f"{iteration}: Waiting to connect to docker container. "
                f"Name: '{self.container_name}', Status: '{container_status}', "
                f"IPAddress: '{self.assigned_ip}', KernelID: '{self.kernel_id}'"
            )

        return container_status

    @overrides
    def terminate_container_resources(self, restart: bool = False) -> None:
        # Remove the container

        result = True  # Since we run containers with remove=True, we'll be optimistic
        container = self._get_container()
        if container:
            try:
                container.remove(force=True)  # Container still exists, attempt forced removal
            except Exception as err:
                self.log.debug(
                    f"Container termination for container: {container.name} raised exception: {err}"
                )
                if isinstance(err, NotFound):
                    pass  # okay if its not found
                else:
                    result = False
                    self.log.warning(f"Error occurred removing container: {err}")

        if result:
            self.log.debug(
                f"{self.__class__.__name__}.terminate_container_resources, "
                f"container {self.container_name}, "
                f"kernel ID: {self.kernel_id} has been terminated."
            )
            self.container_name = None
            result = None  # maintain jupyter contract
        else:
            self.log.warning(
                f"{self.__class__.__name__}.terminate_container_resources, "
                f"container {self.container_name}, "
                f"kernel ID: {self.kernel_id} has not been terminated."
            )
        return result

    def _get_container(self) -> Container:
        # Fetches the container object corresponding the kernel_id label.
        # Only used when docker mode == regular (not swarm)

        container = None
        containers = self.client.containers.list(filters={"label": "kernel_id=" + self.kernel_id})
        num_containers = len(containers)
        if num_containers != 1:
            if num_containers > 1:
                err_msg = (
                    f"{self.__class__.__name__}: Found more than one container "
                    f"({num_containers}) for kernel_id '{self.kernel_id}'!"
                )
                raise RuntimeError(err_msg)
        else:
            container = containers[0]
        return container



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from typing import Optional

from jupyter_client.kernelspec import KernelSpec
from mocks.response_manager import generate_connection_info

from gateway_provisioners.remote_provisioner import RemoteProvisionerBase
from gateway_provisioners.response_manager import ResponseManager

# warning including YarnProvisioner will break the mock of the Yarn ResourceManager

TEST_USER = "test-user"


class ValidatorBase:
    @classmethod
    def create_instance(cls, name: str, seed_env: dict, **kwargs):
        if name == "kubernetes":
            return K8sValidator(name=name, seed_env=seed_env, **kwargs)
        if name == "yarn":
            return YarnValidator(name=name, seed_env=seed_env, **kwargs)
        if name == "docker":
            return DockerValidator(name=name, seed_env=seed_env, **kwargs)
        if name == "docker-swarm":
            return DockerSwarmValidator(name=name, seed_env=seed_env, **kwargs)
        err_msg = f"Invalid name '{name}' encountered!"
        raise AssertionError(err_msg)

    def __init__(self, name: str, seed_env: dict, **kwargs):
        self.name: str = name
        self.seed_env: dict = seed_env
        self.kernel_id: str = kwargs.get("kernel_id")
        self.response_manager: ResponseManager = kwargs.get("response_manager")
        self.kernel_spec: Optional[KernelSpec] = None
        self.provisioner = None

    def validate_provisioner(self, provisioner: RemoteProvisionerBase) -> None:
        assert provisioner.kernel_id == self.kernel_id
        assert provisioner.response_manager == self.response_manager
        assert not provisioner.kernel_username
        assert provisioner.kernel_spec.language == "python"
        self.kernel_spec = provisioner.kernel_spec
        self.provisioner = provisioner

    def validate_pre_launch(self, kwargs: dict) -> None:
        cmd: list = kwargs.get("cmd")
        assert cmd is not None
        assert f"--kernel-id:{self.kernel_id}" in cmd
        assert "--port-range:0..0" in cmd
        assert f"--response-address:{self.response_manager.response_address}" in cmd
        assert f"--public-key:{self.response_manager.public_key}" in cmd

        env: dict = kwargs.get("env")
        assert env is not None
        assert env["KERNEL_ID"] == self.kernel_id
        assert env["KERNEL_USERNAME"] == TEST_USER
        assert env["KERNEL_LANGUAGE"] == self.kernel_spec.language

    def validate_launch_kernel(self, connection_info: dict) -> None:
        assert connection_info == generate_connection_info(self.kernel_id)

    def validate_post_launch(self, kwargs: dict) -> None:
        """Not currently used by GP"""
        pass


class YarnValidator(ValidatorBase):
    """Handles validation of YarnProvisioners"""

    def validate_pre_launch(self, kwargs: dict):
        super().validate_pre_launch(kwargs)

        env: dict = kwargs.get("env")
        assert env["GP_IMPERSONATION_ENABLED"] == "False"
        assert self.provisioner.rm_addr == env["GP_YARN_ENDPOINT"]


class K8sValidator(ValidatorBase):
    """Handles validation of KubernetesProvisioners"""

    def validate_pre_launch(self, kwargs: dict):
        super().validate_pre_launch(kwargs)

        env: dict = kwargs.get("env")
        assert env["KERNEL_UID"] == "1000"
        assert env["KERNEL_GID"] == "100"
        assert env["KERNEL_POD_NAME"] == f"{TEST_USER}-{self.kernel_id}"
        assert env["KERNEL_NAMESPACE"] == "default"
        assert env["KERNEL_SERVICE_ACCOUNT_NAME"] == "default"


class DockerValidator(ValidatorBase):
    """Handles validation of DockerProvisioners"""

    def validate_pre_launch(self, kwargs: dict):
        super().validate_pre_launch(kwargs)

        env: dict = kwargs.get("env")
        assert env["KERNEL_UID"] == "1000"
        assert env["KERNEL_GID"] == "100"
        assert env["GP_DOCKER_NETWORK"] == "bridge"
        assert env["GP_DOCKER_MODE"] == "docker"


class DockerSwarmValidator(ValidatorBase):
    """Handles validation of DockerSwarmProvisioners"""

    def validate_pre_launch(self, kwargs: dict):
        super().validate_pre_launch(kwargs)

        env: dict = kwargs.get("env")
        assert env["KERNEL_UID"] == "1000"
        assert env["KERNEL_GID"] == "100"
        assert env["GP_DOCKER_NETWORK"] == "bridge"
        assert env["GP_DOCKER_MODE"] == "swarm"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Code related to managing kernels running in YARN clusters."""
import asyncio
import errno
import logging
import os
import signal
import socket
import time
from typing import Any, Dict, List, Optional, Tuple

from overrides import overrides
from traitlets import Bool, Unicode, default

try:
    from yarn_api_client.resource_manager import ResourceManager
except ImportError:
    logging.warning(
        "The extra package 'yarn_api_client'is not installed in this environment and is "
        "required.  Ensure that gateway_provisioners is installed by specifying the "
        "extra 'yarn' (e.g., pip install 'gateway_provisioners[yarn]')."
    )
    raise

from .config_mixin import max_poll_attempts, poll_interval
from .remote_provisioner import RemoteProvisionerBase

# Default logging level of yarn-api and underlying connection pool produce too much noise - raise to warning only.
logging.getLogger("yarn_api_client").setLevel(os.getenv("GP_YARN_LOG_LEVEL", logging.WARNING))
logging.getLogger("urllib3.connectionpool").setLevel(
    os.environ.get("GP_YARN_LOG_LEVEL", logging.WARNING)
)

yarn_shutdown_wait_time = float(os.getenv("GP_YARN_SHUTDOWN_WAIT_TIME", "15.0"))
# cert_path: Boolean, defaults to `True`, that controls
#            whether we verify the server's TLS certificate in yarn-api-client.
#            Or a string, in which case it must be a path to a CA bundle(.pem file) to use.
cert_path = os.getenv("GP_YARN_CERT_BUNDLE", True)


class YarnProvisioner(RemoteProvisionerBase):
    """
    Kernel lifecycle management for YARN clusters.
    """

    yarn_endpoint_env = "GP_YARN_ENDPOINT"
    yarn_endpoint = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The http url specifying the YARN Resource Manager. Note: If this value is NOT set,
the YARN library will use the files within the local HADOOP_CONFIG_DIR to determine the
active resource manager. (GP_YARN_ENDPOINT env var)""",
    )

    @default("yarn_endpoint")
    def _yarn_endpoint_default(self):
        return os.getenv(self.yarn_endpoint_env)

    # Alt Yarn endpoint
    alt_yarn_endpoint_env = "GP_ALT_YARN_ENDPOINT"
    alt_yarn_endpoint = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The http url specifying the alternate YARN Resource Manager.  This value should
be set when YARN Resource Managers are configured for high availability.  Note: If both
YARN endpoints are NOT set, the YARN library will use the files within the local
HADOOP_CONFIG_DIR to determine the active resource manager.
(GP_ALT_YARN_ENDPOINT env var)""",
    )

    @default("alt_yarn_endpoint")
    def _alt_yarn_endpoint_default(self):
        return os.getenv(self.alt_yarn_endpoint_env)

    yarn_endpoint_security_enabled_env = "GP_YARN_ENDPOINT_SECURITY_ENABLED"
    yarn_endpoint_security_enabled_default_value = False
    yarn_endpoint_security_enabled = Bool(
        yarn_endpoint_security_enabled_default_value,
        config=True,
        help="""Is YARN Kerberos/SPNEGO Security enabled (True/False).
(GP_YARN_ENDPOINT_SECURITY_ENABLED env var)""",
    )

    @default("yarn_endpoint_security_enabled")
    def _yarn_endpoint_security_enabled_default(self):
        return bool(
            os.getenv(
                self.yarn_endpoint_security_enabled_env,
                self.yarn_endpoint_security_enabled_default_value,
            )
        )

    # Impersonation enabled
    impersonation_enabled_env = "GP_IMPERSONATION_ENABLED"
    impersonation_enabled = Bool(
        False,
        config=True,
        help="""Indicates whether impersonation will be performed during kernel launch.
(GP_IMPERSONATION_ENABLED env var)""",
    )

    @default("impersonation_enabled")
    def _impersonation_enabled_default(self):
        return bool(os.getenv(self.impersonation_enabled_env, "false").lower() == "true")

    initial_states = {"NEW", "SUBMITTED", "ACCEPTED", "RUNNING"}
    final_states = {"FINISHED", "KILLED", "FAILED"}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.application_id = None
        self.last_known_state = None
        self.candidate_queue = None
        self.candidate_partition = None

        # If yarn resource check is enabled, and it isn't available immediately,
        # 20% of kernel_launch_timeout is used to wait
        # and retry at fixed interval before pronouncing as not feasible to launch.
        self.yarn_resource_check_wait_time = 0.20 * self.launch_timeout

    @property
    @overrides
    def has_process(self) -> bool:
        return self.local_proc is not None or self.application_id is not None

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        self.application_id = None
        self.last_known_state = None
        self.candidate_queue = None
        self.candidate_partition = None

        # Transfer impersonation enablement to env.  It is assumed that the kernelspec
        # logic will take the appropriate steps to impersonate the user identified by
        # KERNEL_USERNAME when impersonation_enabled is True.
        env_dict = kwargs.get("env")
        env_dict["GP_IMPERSONATION_ENABLED"] = str(self.impersonation_enabled)

        kwargs = await super().pre_launch(**kwargs)

        self._initialize_resource_manager(**kwargs)

        # checks to see if the queue resource is available
        # if not available, kernel startup is not attempted
        self._confirm_yarn_queue_availability(**kwargs)

        return kwargs

    @overrides
    def get_shutdown_wait_time(self, recommended: Optional[float] = 5.0) -> float:
        # YARN applications tend to take longer than the default 5 second wait time.  Rather than
        # require a command-line option for those using YARN, we'll adjust based on a local env that
        # defaults to 15 seconds.  Note: we'll only adjust if the current wait time is shorter than
        # the desired value.
        if recommended < yarn_shutdown_wait_time:
            recommended = yarn_shutdown_wait_time
            self.log.debug(
                f"{type(self).__name__} shutdown wait time adjusted to {recommended} seconds."
            )

        return recommended

    @overrides
    async def poll(self) -> Optional[int]:
        # Submitting a new kernel/app to YARN will take a while to be ACCEPTED.
        # Thus application ID will probably not be available immediately for poll.
        # So will regard the application as RUNNING when application ID still in ACCEPTED or SUBMITTED state.

        result = 0

        if self._get_application_id():
            state = self._query_app_state_by_id(self.application_id)
            if state in YarnProvisioner.initial_states:
                result = None

        # The following produces too much output (every 3 seconds by default), so commented-out at this time.
        # self.log.debug("YarnProcessProxy.poll, application ID: {}, kernel ID: {}, state: {}".
        #               format(self.application_id, self.kernel_id, state))
        return result

    @overrides
    async def send_signal(self, signum: int) -> None:
        if signum == 0:
            return await self.poll()
        elif signum == signal.SIGKILL:
            return await self.kill()
        else:
            # Yarn api doesn't support the equivalent to interrupts, so take our chances
            # via a remote signal.  Note that this condition cannot check against the
            # signum value because alternate interrupt signals might be in play.
            return await super().send_signal(signum)

    @overrides
    async def kill(self, restart: bool = False) -> None:
        state = None
        result = False
        if self._get_application_id():
            result, state = await self._shutdown_application()

        if result is False:  # We couldn't terminate via Yarn, try remote signal
            result = await super().send_signal(signal.SIGKILL)  # Must use super here, else infinite

        self.log.debug(
            f"YarnProvisioner.kill, application ID: {self.application_id}, "
            f"kernel ID: {self.kernel_id}, state: {state}, result: {result}"
        )
        return result

    @overrides
    async def terminate(self, restart: bool = False) -> None:
        state = None
        result = False
        if self._get_application_id():
            result, state = await self._shutdown_application()

        self.log.debug(
            f"YarnProvisioner.terminate, application ID: {self.application_id}, "
            f"kernel ID: {self.kernel_id}, state: {state}, result: {result}"
        )
        return result

    @overrides
    async def cleanup(self, restart: bool = False) -> None:
        # we might have a defunct process (if using waitAppCompletion = false) - so poll, kill, wait when we have
        # a local_proc.
        if self.local_proc:
            self.log.debug(
                f"YarnProvisioner.cleanup: Clearing possible defunct "
                f"process, pid={self.local_proc.pid}..."
            )

            if self.local_proc.poll():
                self.local_proc.kill()
            self.local_proc.wait()
            self.local_proc = None

        # reset application id to force new query - handles kernel restarts/interrupts
        self.application_id = None

        # for cleanup, we should call the superclass last
        await super().cleanup(restart=restart)

    @overrides
    async def get_provisioner_info(self) -> Dict[str, Any]:
        provisioner_info = await super().get_provisioner_info()
        provisioner_info.update({"application_id": self.application_id})
        return provisioner_info

    @overrides
    async def load_provisioner_info(self, provisioner_info: dict) -> None:
        await super().load_provisioner_info(provisioner_info)
        self.application_id = provisioner_info.get("application_id")

    @overrides
    async def confirm_remote_startup(self) -> None:
        self.start_time = RemoteProvisionerBase.get_current_time()
        i = 0
        ready_to_connect = False  # we're ready to connect when we have a connection file to use
        while not ready_to_connect:
            i += 1
            await self.handle_launch_timeout()

            if self._get_application_id(True):
                # Once we have an application ID, start monitoring state, obtain assigned host and get connection info
                app_state = self._get_application_state()

                if app_state in YarnProvisioner.final_states:
                    error_message = (
                        f"KernelID: '{self.kernel_id}', ApplicationID: '{self.application_id}' "
                        f"unexpectedly found in state '{app_state}' during kernel startup!"
                    )
                    self.log_and_raise(RuntimeError(error_message))

                self.log.debug(
                    f"{i}: State: '{app_state}', Host: '{self.assigned_host}', "
                    f"KernelID: '{self.kernel_id}', ApplicationID: '{self.application_id}'"
                )

                if self.assigned_host != "":
                    ready_to_connect = await self.receive_connection_info()
            else:
                self.detect_launch_failure()

    @overrides
    def log_kernel_launch(self, cmd: List[str]) -> None:
        self.log.info(
            f"{self.__class__.__name__}: kernel launched. YARN RM: {self.rm_addr}, "
            f"pid: {self.local_proc.pid}, Kernel ID: {self.kernel_id}, cmd: '{cmd}'"
        )

    @overrides
    async def handle_launch_timeout(self) -> None:
        """
        Checks to see if the kernel launch timeout has been exceeded while awaiting connection info.

        Note: This is a complete override of the superclass method.
        """
        await asyncio.sleep(poll_interval)
        time_interval = RemoteProvisionerBase.get_time_diff(self.start_time)

        if time_interval > self.launch_timeout:
            reason = (
                f"Application ID is None. Failed to submit a new application to YARN within "
                f"{self.launch_timeout} seconds.  Check server log for more information."
            )

            if self._get_application_id(True):
                if self._query_app_state_by_id(self.application_id) != "RUNNING":
                    reason = (
                        f"YARN resources unavailable after {time_interval} seconds for "
                        f"app {self.application_id}, launch timeout: {self.launch_timeout}!  "
                        "Check YARN configuration."
                    )
                else:
                    reason = (
                        f"App {self.application_id} is RUNNING, but waited too long "
                        f"({self.launch_timeout} secs) to get connection file.  "
                        f"Check YARN logs for more information."
                    )
            await self.kill()
            timeout_message = f"KernelID: '{self.kernel_id}' launch timeout due to: {reason}"
            self.log_and_raise(TimeoutError(timeout_message))

    async def _shutdown_application(self) -> Tuple[Optional[bool], str]:
        """Shuts down the YARN application, returning None if final state is confirmed, False otherwise."""
        result = False
        self._kill_app_by_id(self.application_id)
        # Check that state has moved to a final state (most likely KILLED)
        i = 1
        state = self._query_app_state_by_id(self.application_id)
        while state not in YarnProvisioner.final_states and i <= max_poll_attempts:
            await asyncio.sleep(poll_interval)
            state = self._query_app_state_by_id(self.application_id)
            i += 1

        if state in YarnProvisioner.final_states:
            result = None

        return result, state

    def _confirm_yarn_queue_availability(self, **kwargs: Dict[str, Any]) -> None:
        """
        Submitting jobs to yarn queue and then checking till the jobs are in running state
        will lead to orphan jobs being created in some scenarios.

        We take kernel_launch_timeout time and divide this into two parts.
        If the queue is unavailable we take max 20% of the time to poll the queue periodically
        and if the queue becomes available the rest of timeout is met in 80% of the remaining
        time.

        This algorithm is subject to change. Please read the below cases to understand
        when and how checks are applied.

        Confirms if the yarn queue has capacity to handle the resource requests that
        will be sent to it.

        First check ensures the driver and executor memory request falls within
        the container size of yarn configuration. This check requires executor and
        driver memory to be available in the env.

        Second,Current version of check, takes into consideration node label partitioning
        on given queues. Provided the queue name and node label this checks if
        the given partition has capacity available for kernel startup.

        All Checks are optional. If we have KERNEL_EXECUTOR_MEMORY and KERNEL_DRIVER_MEMORY
        specified, first check is performed.

        If we have KERNEL_QUEUE and KERNEL_NODE_LABEL specified, second check is performed.

        Proper error messages are sent back for user experience
        :param kwargs:
        :return:
        """
        env_dict = kwargs.get("env", {})

        executor_memory = int(env_dict.get("KERNEL_EXECUTOR_MEMORY", 0))
        driver_memory = int(env_dict.get("KERNEL_DRIVER_MEMORY", 0))

        if executor_memory * driver_memory > 0:
            container_memory = self.resource_mgr.cluster_node_container_memory()
            if max(executor_memory, driver_memory) > container_memory:
                self.log_and_raise(
                    ValueError("Container Memory not sufficient for a executor/driver allocation")
                )

        candidate_queue_name = env_dict.get("KERNEL_QUEUE", None)
        node_label = env_dict.get("KERNEL_NODE_LABEL", None)
        partition_availability_threshold = float(env_dict.get("YARN_PARTITION_THRESHOLD", 95.0))

        if candidate_queue_name is None or node_label is None:
            return

        # else the resources may or may not be available now. it may be possible that if we wait then the resources
        # become available. start  a timeout process

        self.start_time = RemoteProvisionerBase.get_current_time()
        self.candidate_queue = self.resource_mgr.cluster_scheduler_queue(candidate_queue_name)

        if self.candidate_queue is None:
            self.log.warning(
                f"Queue: {candidate_queue_name} not found in cluster.  "
                "Availability check will not be performed"
            )
            return

        self.candidate_partition = self.resource_mgr.cluster_queue_partition(
            self.candidate_queue, node_label
        )

        if self.candidate_partition is None:
            self.log.debug(
                f"Partition: {node_label} not found in {candidate_queue_name} queue."
                "Availability check will not be performed"
            )
            return

        self.log.debug(
            f"Checking endpoint: {self.yarn_endpoint} if partition: {self.candidate_partition} "
            f"has used capacity <= {partition_availability_threshold}%"
        )

        yarn_available = self.resource_mgr.cluster_scheduler_queue_availability(
            self.candidate_partition, partition_availability_threshold
        )
        if not yarn_available:
            self.log.debug(
                f"Retrying for {self.yarn_resource_check_wait_time} ms since resources are not available"
            )
            while not yarn_available:
                self._handle_yarn_queue_timeout()
                yarn_available = self.resource_mgr.cluster_scheduler_queue_availability(
                    self.candidate_partition, partition_availability_threshold
                )

        # subtracting the total amount of time spent for polling for queue availability
        self.launch_timeout -= RemoteProvisionerBase.get_time_diff(self.start_time)

    def _handle_yarn_queue_timeout(self) -> None:
        time.sleep(poll_interval)
        time_interval = RemoteProvisionerBase.get_time_diff(self.start_time)

        if time_interval > self.yarn_resource_check_wait_time:
            reason = f"Yarn Compute Resource is unavailable after {self.yarn_resource_check_wait_time} seconds"
            self.log_and_raise(TimeoutError(reason))

    def _initialize_resource_manager(self, **kwargs: Optional[Dict[str, Any]]) -> None:
        """Initialize the Hadoop YARN Resource Manager instance used for this kernel's lifecycle."""

        endpoints = None
        if self.yarn_endpoint:
            endpoints = [self.yarn_endpoint]

            # Only check alternate if "primary" is set.
            if self.alt_yarn_endpoint:
                endpoints.append(self.alt_yarn_endpoint)

        if self.yarn_endpoint_security_enabled:
            from requests_kerberos import HTTPKerberosAuth

            auth = HTTPKerberosAuth()
        else:
            # If we have the appropriate version of yarn-api-client, use its SimpleAuth class.
            # This allows EG to continue to issue requests against the YARN api when anonymous
            # access is not allowed. (Default is to allow anonymous access.)
            try:
                from yarn_api_client.auth import SimpleAuth

                auth = SimpleAuth(self.kernel_username)
                self.log.debug(
                    f"Using SimpleAuth with '{self.kernel_username}' against endpoints: {endpoints}"
                )
            except ImportError:
                auth = None

        self.resource_mgr = ResourceManager(
            service_endpoints=endpoints, auth=auth, verify=cert_path
        )

        self.rm_addr = self.resource_mgr.get_active_endpoint()

    def _get_application_state(self) -> str:
        """
        Gets the current application state using the application_id already obtained.

        Once the assigned host has been identified, 'amHostHttpAddress' is no longer accessed.
        """
        app_state = self.last_known_state
        app = self._query_app_by_id(self.application_id)
        if app:
            if app.get("state"):
                app_state = app.get("state")
                self.last_known_state = app_state

            if self.assigned_host == "" and app.get("amHostHttpAddress"):
                self.assigned_host = app.get("amHostHttpAddress").split(":")[0]
                # Set the assigned ip to the actual host where the application landed.
                self.assigned_ip = socket.gethostbyname(self.assigned_host)

        return app_state

    def _get_application_id(self, ignore_final_states: bool = False) -> Optional[str]:
        """
        Return the kernel's YARN application ID if available, otherwise None.

        If we're obtaining application_id from scratch, do not consider kernels in final states.
        :param ignore_final_states:
        :returns Optional[str] - the YARN application ID or None if not available
        """
        if not self.application_id:
            app = self._query_app_by_name(self.kernel_id)
            state_condition = True
            if type(app) is dict:
                state = app.get("state")
                self.last_known_state = state

                if ignore_final_states:
                    state_condition = state not in YarnProvisioner.final_states

                if len(app.get("id", "")) > 0 and state_condition:
                    self.application_id = app["id"]
                    time_interval = RemoteProvisionerBase.get_time_diff(self.start_time)
                    self.log.info(
                        f"ApplicationID: '{app['id']}' assigned for KernelID: '{self.kernel_id}', "
                        f"state: {state}, {time_interval} seconds after starting."
                    )
            if not self.application_id:
                self.log.debug(
                    f"ApplicationID not yet assigned for KernelID: '{self.kernel_id}' - retrying..."
                )
        return self.application_id

    def _query_app_by_name(self, kernel_id: str) -> Optional[dict]:
        """
        Retrieve application by using kernel_id as the unique app name.

        With the started_time_begin as a parameter to filter applications started earlier than the target one from YARN.
        When submit a new app, it may take a while for YARN to accept and run and generate the application ID.
        Note: if a kernel restarts with the same kernel id as app name, multiple applications will be returned.
        For now, the app/kernel with the top most application ID will be returned as the target app, assuming the app
        ID will be incremented automatically on the YARN side.

        :param kernel_id: as the unique app name for query
        :return: The JSON object of an application or None on failure
        """
        top_most_app_id = ""
        target_app = None
        try:
            response = self.resource_mgr.cluster_applications(
                started_time_begin=str(self.start_time)
            )
        except OSError as sock_err:
            if sock_err.errno == errno.ECONNREFUSED:
                self.log.warning(
                    f"YARN RM address: '{self.rm_addr}' refused the connection.  "
                    f"Is the resource manager running?"
                )
            else:
                self.log.warning(
                    f"Query for kernel ID '{kernel_id}' failed with exception: "
                    f"{type(sock_err)} - '{sock_err}'.  Continuing..."
                )
        except Exception as e:
            self.log.warning(
                f"Query for kernel ID '{kernel_id}' failed with exception: "
                f"{type(e)} - '{e}'.  Continuing..."
            )
        else:
            data = response.data
            if type(data) is dict and type(data.get("apps")) is dict and "app" in data.get("apps"):
                for app in data["apps"]["app"]:
                    if app.get("name", "").find(kernel_id) >= 0 and app.get("id") > top_most_app_id:
                        target_app = app
                        top_most_app_id = app.get("id")
        return target_app

    def _query_app_by_id(self, app_id: str) -> Optional[dict]:
        """Retrieve an application by application ID.

        :param app_id
        :return: The JSON object of an application or None on failure.
        """
        app = None
        try:
            response = self.resource_mgr.cluster_application(application_id=app_id)
        except Exception as e:
            self.log.warning(
                f"Query for application ID '{app_id}' failed with exception: '{e}'.  Continuing..."
            )
        else:
            data = response.data
            if type(data) is dict and "app" in data:
                app = data["app"]

        return app

    def _query_app_state_by_id(self, app_id: str) -> str:
        """Return the state of an application. If a failure occurs, the last known state is returned.

        :param app_id:
        :return: application state (str)
        """
        state = self.last_known_state
        try:
            response = self.resource_mgr.cluster_application_state(application_id=app_id)
        except Exception as e:
            self.log.warning(
                f"Query for application '{app_id}' state failed with exception: '{e}'.  "
                f"Continuing with last known state = '{state}'..."
            )
        else:
            state = response.data["state"]
            self.last_known_state = state

        return state

    def _kill_app_by_id(self, app_id: str) -> dict:
        """Kill an application. If the app's state is FINISHED or FAILED, it won't be changed to KILLED.

        :param app_id
        :return: The JSON response of killing the application.
        """

        response = {}
        try:
            response = self.resource_mgr.cluster_application_kill(application_id=app_id)
        except Exception as e:
            self.log.warning(
                f"Termination of application '{app_id}' failed with exception: '{e}'.  Continuing..."
            )
        return response



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import os
import shutil
from string import Template
from typing import Any

from jupyter_core.application import JupyterApp
from overrides import overrides
from traitlets import Bool, List, TraitError, Unicode, validate

from .._version import __version__
from .base_app import (
    PYTHON,
    SCALA,
    SUPPORTED_LANGUAGES,
    BaseApp,
    R,
    kernel_launchers_dir,
)

LANG_DIR_NAMES = {PYTHON: "python", SCALA: "scala", R: "R"}
BOOTSTRAP_FILE_NAME = "bootstrap-kernel.sh"
# Note that BOOTSTRAP_DIR is not configurable due to its reference buries in the
# sparkoperator.k8s.io/v1beta2.yaml.j2 template used by the Spark Operator.  We can
# revisit how to make this configurable later.
BOOTSTRAP_DIR = "/usr/local/bin"


class ImageBootstrapInstaller(BaseApp):
    """CLI for extension management."""

    name = "jupyter-image-bootstrap"
    description = (
        "Installs the bootstrap script and kernel launchers for use within a kernel "
        "image used by Remote Provisioners."
    )
    # Note that the left justification of the second example is necessary to ensure proper
    # alignment with the first example during --help output.
    examples = """
    jupyter-image-bootstrap install --languages=Python --languages=R

jupyter-image-bootstrap install --languages=Python --languages=Scala
    """

    languages = List(
        Unicode(),
        ["Python"],
        config=True,
        help="""The languages corresponding to the kernel-launchers to install into the kernel image.
    All values must be one of 'Python', 'R', or 'Scala'.""",
    )

    @validate("languages")
    def _languages_validate(self, proposal: dict[str, Any]) -> List:
        value = proposal["value"]
        try:
            for lang in value:
                assert lang.lower() in SUPPORTED_LANGUAGES
        except AssertionError as ae:
            err_msg = (
                f"Invalid languages value {value}, at least one of which is "
                f"not in {SUPPORTED_LANGUAGES} (case-insensitive)"
            )
            raise TraitError(err_msg) from ae
        return value

    launchers_only = Bool(
        False, config=True, help="Only install kernel launchers, no bootstrap script."
    )

    aliases = {
        "languages": "ImageBootstrapInstaller.languages",
    }
    aliases.update(BaseApp.aliases)

    flags = {
        "launchers-only": (
            {"ImageBootstrapInstaller": {"launchers_only": True}},
            "Only install kernel launchers, no bootstrap script.",
        ),
    }
    flags.update(BaseApp.flags)

    kernel_spec_install = False  # Set to false when bootstrap installation occurs

    @overrides
    def detect_missing_extras(self):
        for lang in self.languages:
            if lang.lower() == SCALA:
                self._detect_missing_toree_jar()
            elif lang.lower() == R:
                self._detect_missing_rscript()

    @overrides
    def install_files(self):
        """Sets up the bootstrap-kernel.sh and corresponding kernel-launchers for use in kernel images"""

        parent_dir = os.path.join(BOOTSTRAP_DIR, "kernel-launchers")
        for lang in self.languages:
            lang_dir_name = LANG_DIR_NAMES.get(lang.lower())
            target_dir = os.path.join(parent_dir, lang_dir_name)
            self._copy_launcher_files(lang_dir_name, target_dir)
            if lang_dir_name == SCALA and not self.toree_jar_path:
                # If we're installing a scala kernel launcher and don't have the toree jar file, issue
                # a warning indicating that the scala kernel needs that file in its kernel-launcher
                # directory hierarchy.
                kspec_toree_jar_location = os.path.join(target_dir, "lib")
                self.log.warning(
                    "The Apache Toree kernel is either not installed or it's jar file cannot be determined. "
                    f"Please ensure that the Toree jar file is placed into '{kspec_toree_jar_location}' "
                    f"prior to using the kernel image for scala."
                )
        if self.languages:
            self.log.info(
                f"Kernel-launcher files have been copied to {parent_dir} "
                f"for the following languages: {self.languages}."
            )

        if not self.launchers_only:
            bootstrap_file = os.path.join(kernel_launchers_dir, "bootstrap", BOOTSTRAP_FILE_NAME)
            target_bootstrap_file = os.path.join(BOOTSTRAP_DIR, BOOTSTRAP_FILE_NAME)
            shutil.copyfile(bootstrap_file, target_bootstrap_file)
            self._finalize_bootstrap(target_bootstrap_file)
            self.log.info(f"{BOOTSTRAP_FILE_NAME} has been copied to {BOOTSTRAP_DIR}.")
            self.log.info(
                f"The CMD entry in the Dockerfile should be updated to: CMD {target_bootstrap_file}"
            )

    @staticmethod
    def _finalize_bootstrap(bootstrap_file: str):
        subs = {"install_dir": BOOTSTRAP_DIR}
        bootstrap_str = ""
        with open(bootstrap_file) as f:
            for line in f:
                bootstrap_str = bootstrap_str + line
            f.close()
        post_subs = Template(bootstrap_str).safe_substitute(subs)
        with open(bootstrap_file, "w+") as f:
            f.write(post_subs)


class ImageBootstrapApp(JupyterApp):
    """Application responsible for driving the creation of Kubernetes-based kernel specifications."""

    version = __version__
    name = "jupyter image-bootstrap"
    description = """Application used to bootstrap kernel images with the appropriate
kernel launchers for use by Remote Provisioners."""
    subcommands = dict(
        {
            "install": (
                ImageBootstrapInstaller,
                ImageBootstrapInstaller.description.splitlines()[0],
            ),
        }
    )
    aliases = {}
    flags = {}

    def start(self):
        super().start()

        if self.subapp is None:
            print(f"No subcommand specified. Must specify one of: {list(self.subcommands)}")
            print()
            self.print_description()
            self.print_subcommands()
            self.exit(1)


if __name__ == "__main__":
    ImageBootstrapInstaller.launch_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import argparse
import logging
import os
import signal
import tempfile
from collections.abc import Callable
from threading import Thread
from typing import Any, Optional

from server_listener import setup_server_listener

LAUNCHER_VERSION = 1  # Indicate to server the version of this launcher (payloads may vary)

# Minimum port range size and max retries
min_port_range_size = int(os.getenv("MIN_PORT_RANGE_SIZE", "1000"))

log_level = os.getenv("LOG_LEVEL", "10")
log_level = int(log_level) if log_level.isdigit() else log_level

logging.basicConfig(format="[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s] %(message)s")

# Protect the instance with a dunder so it isn't remove from the namespace
__logger = logging.getLogger("launch_ipykernel")
__logger.setLevel(log_level)

DEFAULT_KERNEL_CLASS_NAME = "ipykernel.ipkernel.IPythonKernel"
__spark_context = None


class ExceptionThread(Thread):
    # Wrap thread to handle the exception
    def __init__(self, target: Callable):
        self.target = target
        self.exc = None
        Thread.__init__(self)

    def run(self):
        try:
            self.target()
        except Exception as exc:
            self.exc = exc


def initialize_namespace(namespace: dict, cluster_type: str = "spark") -> None:
    """Initialize the kernel namespace.

    Parameters
    ----------
    namespace : Dict
        The namespace to initialize
    cluster_type : {'spark', 'dask', 'none'}
        The cluster type to initialize. ``'none'`` results in no variables in
        the initial namespace.
    """
    if cluster_type == "spark":
        try:
            from pyspark.sql import SparkSession
        except ImportError:
            __logger.info(
                "A spark context was desired but the pyspark distribution is not present.  "
                "Spark context creation will not occur."
            )
            return

        def initialize_spark_session():
            import atexit

            """Initialize Spark session and replace global variable
            placeholders with real Spark session object references."""
            spark = SparkSession.builder.getOrCreate()

            global __spark_context
            __spark_context = spark.sparkContext

            # Stop the spark session on exit
            atexit.register(lambda: spark.stop())

            namespace.update(
                {
                    "spark": spark,
                    "sc": spark.sparkContext,
                    "sql": spark.sql,
                }
            )

        init_thread = ExceptionThread(target=initialize_spark_session)
        spark = WaitingForSparkSessionToBeInitialized("spark", init_thread, namespace)
        sc = WaitingForSparkSessionToBeInitialized("sc", init_thread, namespace)

        def sql(query):
            """Placeholder function. When called will wait for Spark session to be
            initialized and call ``spark.sql(query)``"""
            return spark.sql(query)

        namespace.update({"spark": spark, "sc": sc, "sql": sql})

        init_thread.start()

    elif cluster_type == "dask":
        import dask_yarn

        cluster = dask_yarn.YarnCluster.from_current()
        namespace.update({"cluster": cluster})
    elif cluster_type != "none":
        raise RuntimeError("Unknown cluster_type: %r" % cluster_type)


class WaitingForSparkSessionToBeInitialized:
    """Wrapper object for SparkContext and other Spark session variables while the real Spark session is being
    initialized in a background thread. The class name is intentionally worded verbosely explicit as it will show up
    when executing a cell that contains only a Spark session variable like ``sc``.
    """

    # private and public attributes that show up for tab completion,
    # to indicate pending initialization of Spark session
    _WAITING_FOR_SPARK_SESSION_TO_BE_INITIALIZED = "Spark Session not yet initialized ..."
    WAITING_FOR_SPARK_SESSION_TO_BE_INITIALIZED = "Spark Session not yet initialized ..."

    # the same wrapper class is used for all Spark session variables, so we need to record the name of the variable
    def __init__(self, global_variable_name: str, init_thread: Thread, namespace: dict):
        self._spark_session_variable = global_variable_name
        self._init_thread = init_thread
        self._namespace = namespace

    # we intercept all method and attribute references on our temporary Spark session variable,
    # wait for the thread to complete initializing the Spark sessions and then we forward the
    # call to the real Spark objects
    def __getattr__(self, name: str):
        # ignore tab-completion request for __members__ or __methods__ and ignore meta property requests
        if name.startswith("__") or name.startswith("_ipython_") or name.startswith("_repr_"):
            return
        else:
            # wait on thread to initialize the Spark session variables in global variable scope
            self._init_thread.join(timeout=None)
            exc = self._init_thread.exc
            if exc:
                err_msg = f"Variable: {self._spark_session_variable} was not initialized properly."
                raise RuntimeError(err_msg) from exc

            # now return attribute/function reference from actual Spark object
            return getattr(self._namespace[self._spark_session_variable], name)


def _validate_port_range(port_range: Optional[str]) -> tuple[int, int]:
    # if no argument was provided, return a range of 0
    if not port_range:
        return 0, 0

    try:
        port_ranges = port_range.split("..")
        lower_port = int(port_ranges[0])
        upper_port = int(port_ranges[1])

        port_range_size = upper_port - lower_port
        if port_range_size != 0:
            if port_range_size < min_port_range_size:
                err_msg = (
                    f"Port range validation failed for range: '{port_range}'.  Range size must be at least "
                    f"{min_port_range_size} as specified by env MIN_PORT_RANGE_SIZE"
                )
                raise RuntimeError(err_msg)
    except ValueError as ve:
        err_msg = f"Port range validation failed for range: '{port_range}'.  Error was: {ve}"
        raise RuntimeError(err_msg) from ve
    except IndexError as ie:
        err_msg = f"Port range validation failed for range: '{port_range}'.  Error was: {ie}"
        raise RuntimeError(err_msg) from ie

    return lower_port, upper_port


def determine_connection_file(kid: str) -> str:
    # Create a temporary (and empty) file using kernel-id
    fd, conn_file = tempfile.mkstemp(suffix=".json", prefix=f"kernel-{kid}_")
    os.close(fd)
    __logger.debug(f"Using connection file '{conn_file}'.")

    return conn_file


def cancel_spark_jobs(sig: int, frame: Any) -> None:
    if __spark_context is None:
        return
    try:
        __spark_context.cancelAllJobs()
    except Exception as e:
        if e.__class__.__name__ == "Py4JError":
            try:
                __spark_context.cancelAllJobs()
            except Exception as ex:
                __logger.error(
                    f"Error occurred while re-attempting Spark job cancellation when interrupting the kernel: {ex}"
                )
        else:
            __logger.error(
                f"Error occurred while attempting Spark job cancellation when interrupting the kernel: {e}"
            )


def import_item(name: str) -> Any:
    """Import and return ``bar`` given the string ``foo.bar``.
    Calling ``bar = import_item("foo.bar")`` is the functional equivalent of
    executing the code ``from foo import bar``.
    Parameters
    ----------
    name : string
      The fully qualified name of the module/package being imported.
    Returns
    -------
    mod : module object
       The module that was imported.
    """

    parts = name.rsplit(".", 1)
    if len(parts) == 2:
        # called with 'foo.bar....'
        package, obj = parts
        module = __import__(package, fromlist=[obj])
        try:
            pak = getattr(module, obj)
        except AttributeError as ae:
            err_msg = f"No module named '{obj}'"
            raise ImportError(err_msg) from ae
        return pak
    else:
        # called with un-dotted string
        return __import__(parts[0])


def start_ipython(
    namespace: dict,
    cluster_type: str = "spark",
    kernel_class_name: str = DEFAULT_KERNEL_CLASS_NAME,
    **kwargs: Any,
) -> None:
    from ipykernel.kernelapp import IPKernelApp

    # Capture the kernel class before removing 'import_item' from the namespace
    kernel_class = import_item(kernel_class_name)

    # create an initial list of variables to clear
    # we do this without deleting to preserve the locals so that
    # initialize_namespace isn't affected by this mutation
    to_delete = [k for k in namespace if not k.startswith("__")]

    # initialize the namespace with the proper variables
    initialize_namespace(namespace, cluster_type=cluster_type)

    # delete the extraneous variables
    for k in to_delete:
        del namespace[k]

    # Start the kernel.
    app = IPKernelApp.instance(kernel_class=kernel_class, user_ns=namespace, **kwargs)
    app.initialize([])
    app.start()

    # cleanup
    conn_file = kwargs["connection_file"]
    __logger.info(f"IPKernelApp has terminated, removing connection file '{conn_file}'.")
    try:
        import os  # re-import os since it's removed during namespace manipulation during startup

        os.remove(conn_file)
    except Exception as e:
        __logger.error(f"Could not delete connection file '{conn_file}' at exit due to error: {e}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--response-address",
        dest="response_address",
        nargs="?",
        metavar="<ip>:<port>",
        help="Connection address (<ip>:<port>) for returning connection file",
    )
    parser.add_argument(
        "--kernel-id",
        dest="kernel_id",
        nargs="?",
        help="Indicates the id associated with the launched kernel.",
    )
    parser.add_argument(
        "--public-key",
        dest="public_key",
        nargs="?",
        help="Public key used to encrypt connection information",
    )
    parser.add_argument(
        "--port-range",
        dest="port_range",
        nargs="?",
        metavar="<lowerPort>..<upperPort>",
        help="Port range to impose for kernel ports",
    )
    parser.add_argument(
        "--spark-context-initialization-mode",
        dest="init_mode",
        nargs="?",
        default="none",
        help="the initialization mode of the spark context: lazy, eager or none",
    )
    parser.add_argument(
        "--cluster-type",
        dest="cluster_type",
        nargs="?",
        default="spark",
        help="the kind of cluster to initialize: spark, dask, or none",
    )
    parser.add_argument(
        "--kernel-class-name",
        dest="kernel_class_name",
        nargs="?",
        default=DEFAULT_KERNEL_CLASS_NAME,
        help="Indicates the name of the kernel class to use.  Must be a subclass of 'ipykernel.kernelbase.Kernel'.",
    )

    arguments = vars(parser.parse_args())
    response_addr = arguments["response_address"]
    kernel_id = arguments["kernel_id"]
    public_key = arguments["public_key"]
    lower_port, upper_port = _validate_port_range(arguments["port_range"])
    spark_init_mode = arguments["init_mode"]
    cluster_type = arguments["cluster_type"]
    kernel_class_name = arguments["kernel_class_name"]
    ip = "0.0.0.0"  # noqa: S104

    if kernel_id is None:
        err_msg = "Parameter '--kernel-id' must be provided!"
        raise RuntimeError(err_msg)

    if response_addr is None:
        err_msg = "Parameter '--response-address' must be provided!"
        raise RuntimeError(err_msg)

    if public_key is None:
        err_msg = "Parameter '--public-key' must be provided!"
        raise RuntimeError(err_msg)

    # Initialize the kernel namespace for the given cluster type
    if cluster_type == "spark" and spark_init_mode == "none":
        cluster_type = "none"

    # Create connection information
    connection_file = determine_connection_file(kernel_id)

    setup_server_listener(
        connection_file,
        os.getpid(),
        lower_port,
        upper_port,
        response_addr,
        kernel_id,
        public_key,
        cluster_type,
    )

    # setup sig handler to cancel spark jobs on interrupts
    if cluster_type == "spark":
        signal.signal(signal.SIGUSR2, cancel_spark_jobs)

    # launch the IPython kernel instance
    start_ipython(
        locals(),
        cluster_type=cluster_type,
        connection_file=connection_file,
        ip=ip,
        kernel_class_name=kernel_class_name,
    )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

response_manager_registration = {}


def generate_connection_info(id: str) -> dict:
    comm_port = hash(id) % 65535
    return {
        "comm_port": comm_port,
        "shell_port": comm_port + 1,
        "iopub_port": comm_port + 2,
        "stdin_port": comm_port + 3,
        "control_port": comm_port + 4,
        "hb_port": comm_port + 5,
        "ip": "127.0.0.1",
        "key": id.replace("-", ""),
        "transport": "tcp",
        "signature_scheme": "hmac-sha256",
        "kernel_name": "python3",
    }


# This avoids the response manager from listening for requests and
# an annoying prompt in debug mode.
def mock_socket_listen(self, backlog: int) -> None:
    pass


def mock_register_event(self, kernel_id: str) -> None:
    assert kernel_id not in response_manager_registration
    response_manager_registration[kernel_id] = {}


async def mock_get_connection_info(self, kernel_id: str) -> dict:
    assert kernel_id in response_manager_registration
    return generate_connection_info(kernel_id)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


class DockerResource:
    """Track the current state of the resource"""

    def __init__(self, env: dict):
        self.env = env
        self.kernel_id: str = env.get("KERNEL_ID")
        self.kernel_username: str = env.get("KERNEL_USERNAME")
        self.container_name = f"{self.kernel_username}-{self.kernel_id}"
        self.docker_network = env.get("GP_DOCKER_NETWORK")
        self.is_swarm = env.get("GP_DOCKER_MODE") == "swarm"
        self.status: str = "created"
        if self.is_swarm:
            self.status = "preparing"
        self.query_counter: int = 1


docker_resources: dict = {}


class MockService:
    def __init__(self, resource: DockerResource):
        self.resource = resource
        self.name = resource.container_name
        self.status = resource.status
        self.attrs = {
            "NetworkSettings": {
                "IPAddress": "127.0.0.1",
                "Networks": {resource.docker_network: {"IPAddress": "127.0.0.1"}},
            }
        }
        task = {
            "ID": hash(self.resource.kernel_id),
            "Status": {"State": self.resource.status},
            "NetworksAttachments": [{"Addresses": ["127.0.0.1/xxx"]}],
        }
        self.task_list: list = [task]

    def tasks(self, **kwargs):
        return self.task_list

    def remove(self, **kwargs):
        docker_resources.pop(self.resource.kernel_id)


class MockServiceCollection:  # (ServiceCollection):
    def __init__(self, **kwargs):
        pass

    def list(self, **kwargs):
        """Get a collection of Containers"""
        # This will be called with a filters object, the "label" key of
        # which contains the kernel_id, so we need to pick that out to
        # locate the appropriate entry:
        # {"label": "kernel_id=" + self.kernel_id}"

        services = []
        label = kwargs.get("filters", {}).get("label", "")
        assert label, "No label found in filters - can't list containers!"
        kernel_id = label.split("=")[1]
        if kernel_id in docker_resources:
            resource = docker_resources.get(kernel_id)
            if resource.query_counter >= 3:  # time to return
                resource.status = "running"
                service = MockService(resource)
                services.append(service)
            else:
                resource.status = "starting"
            resource.query_counter += 1

        return services


class MockContainer:
    def __init__(self, resource: DockerResource):
        self.resource = resource
        self.name = resource.container_name
        self.status = resource.status
        self.attrs = {
            "NetworkSettings": {
                "IPAddress": "127.0.0.1",
                "Networks": {resource.docker_network: {"IPAddress": "127.0.0.1"}},
            }
        }

    def remove(self, **kwargs):
        docker_resources.pop(self.resource.kernel_id)


class MockContainerCollection:  # (ContainerCollection):
    def __init__(self, **kwargs):
        pass

    def list(self, **kwargs):
        """Get a collection of Containers"""
        # This will be called with a filters object, the "label" key of
        # which contains the kernel_id, so we need to pick that out to
        # locate the appropriate entry:
        # {"label": "kernel_id=" + self.kernel_id}"

        containers = []
        label = kwargs.get("filters", {}).get("label", "")
        assert label, "No label found in filters - can't list containers!"
        kernel_id = label.split("=")[1]
        if kernel_id in docker_resources:
            resource = docker_resources.get(kernel_id)
            if resource.query_counter >= 3:  # time to return
                resource.status = "running"
                container = MockContainer(resource)
                containers.append(container)
            resource.query_counter += 1

        return containers


class MockDockerClient:  # (DockerClient):
    def __init__(self, **kwargs):
        pass

    @property
    def containers(self):
        return MockContainerCollection(client=self)

    @property
    def services(self):
        return MockServiceCollection(client=self)


def mock_docker_from_env():  # -> DockerClient:
    return MockDockerClient()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import os

from jupyter_core.application import JupyterApp
from overrides import overrides
from traitlets import Bool, Unicode, default

from .._version import __version__
from .base_app import DEFAULT_LANGUAGE, PYTHON, SCALA, BaseSpecSparkApp, R

SPARK_OP = "spark_operator"

KERNEL_SPEC_TEMPLATE_NAMES = {
    PYTHON: "container_python",
    SCALA: "container_scala",
    R: "container_r",
    SPARK_OP: "k8s_python_spark_operator",
}
DEFAULT_KERNEL_NAMES = {
    PYTHON: "k8s_python",
    SCALA: "k8s_scala",
    R: "k8s_r",
    SPARK_OP: "k8s_python_spark_operator",
}
DEFAULT_DISPLAY_NAMES = {
    PYTHON: "Kubernetes Python",
    SCALA: "Kubernetes Scala",
    R: "Kubernetes R",
    SPARK_OP: "Kubernetes Spark Operator",
}
DEFAULT_IMAGE_NAMES = {
    PYTHON: "elyra/kernel-py",
    SCALA: "elyra/kernel-scala",
    R: "elyra/kernel-r",
}
DEFAULT_SPARK_IMAGE_NAMES = {
    PYTHON: "elyra/kernel-spark-py",
    SCALA: "elyra/kernel-scala",
    R: "elyra/kernel-spark-r",
}

SPARK_SUFFIX = "_spark"
SPARK_DISPLAY_NAME_SUFFIX = " (with Spark)"

PROVISIONER_NAME = "kubernetes-provisioner"
SPARK_OP_PROVISIONER_NAME = "spark-operator-provisioner"
LAUNCHER_NAME = "launch_kubernetes.py"
SPARK_OP_LAUNCHER_NAME = "launch_custom_resource.py"


class K8sSpecInstaller(BaseSpecSparkApp):
    """CLI for extension management."""

    name = "jupyter-k8s-spec"
    description = "Creates a Jupyter kernel specification for use within a Kubernetes cluster."
    # Note that the left justification of the second example is necessary to ensure proper
    # alignment with the first example during --help output.
    examples = """
    jupyter-k8s-spec install --language=R --kernel-name=r_k8s --image-name=foo/my_r_kernel_image:v4_0

jupyter-k8s-spec install --language=Scala --spark --kernel-name=scala_k8s_spark --display-name='Scala on Kubernetes with Spark'

jupyter-k8s-spec install --spark --crd --display-name='Python SparkOperator"
    """

    @default("kernel_name")
    def _kernel_name_default(self) -> str:
        return DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]

    @default("display_name")
    def _display_name_default(self) -> str:
        return DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]

    # Image name
    image_name_env = "GP_IMAGE_NAME"
    image_name = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The kernel image to use for this kernel specification. If this specification is
enabled for Spark usage, this image will be the driver image. (GP_IMAGE_NAME env var)""",
    )

    @default("image_name")
    def _image_name_default(self):
        return os.getenv(self.image_name_env)

    executor_image_name_env = "GP_EXECUTOR_IMAGE_NAME"
    executor_image_name = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The executor image to use for this kernel specification.  Only applies to
Spark-enabled kernel specifications.  (GP_EXECUTOR_IMAGE_NAME env var)""",
    )

    @default("executor_image_name")
    def _executor_image_name_default(self):
        return os.getenv(self.executor_image_name_env)

    # Flags
    tensorflow = Bool(False, config=True, help="""Install kernelspec for use with Tensorflow.""")
    crd = Bool(
        False,
        config=True,
        help="""Install kernelspec for use with a Custom Resource Definition.  When combined with --spark,
will configure the SparkOperatorProvisioner for Spark Application CRDs.""",
    )
    spark = Bool(
        False,
        config=True,
        help="""Install kernelspec for use with Spark.  When combined with --crd,
will configure the SparkOperatorProvisioner for Spark Application CRDs.""",
    )
    provisioner_name = Unicode(PROVISIONER_NAME, config=False)
    launcher_name = Unicode(LAUNCHER_NAME, config=False)

    aliases = {
        "image-name": "K8sSpecInstaller.image_name",
        "executor-image-name": "K8sSpecInstaller.executor_image_name",
    }
    aliases.update(BaseSpecSparkApp.aliases)

    flags = {}
    flags.update(BaseSpecSparkApp.flags)
    flags.update(
        {
            "tensorflow": (
                {"K8sSpecInstaller": {"tensorflow": True}},
                tensorflow.help,
            ),
            "crd": (
                {"K8sSpecInstaller": {"crd": True}},
                crd.help,
            ),
            # Override the spark flag so this help string is present.
            "spark": (
                {"K8sSpecInstaller": {"spark": True}},
                spark.help,
            ),
        }
    )

    @overrides
    def detect_missing_extras(self):
        super().detect_missing_extras()
        try:
            import jinja2  # noqa: F401
            import kubernetes  # noqa: F401
        except ImportError:
            self.log.warning(
                "At least one of the extra packages 'kubernetes' or 'jinja2' are not installed in "
                "this environment and are required.  Ensure that gateway_provisioners is installed "
                "by specifying the extra 'k8s' (e.g., pip install 'gateway_provisioners[k8s]')."
            )

    @overrides
    def validate_parameters(self):
        super().validate_parameters()

        entered_language = self.language
        self.language = self.language.lower()
        self.launcher_dir_name = "kubernetes"
        self.resource_dir_name = self.language

        if self.crd is True:
            if self.spark is False:
                reason = (
                    "--crd requires a specifying option (e.g., --spark) to determine which "
                    "CustomResourceProvisioner to configure."
                )
                raise RuntimeError(reason)
            if self.language != PYTHON:
                self.log.warning(
                    f"CRD support only works with Python, changing language from {entered_language} to Python."
                )
                self.language = PYTHON
            # if kernel and display names are still defaulted, silently convert to lang default and append spark suffix
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[SPARK_OP]
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[SPARK_OP]

            self.kernel_spec_dir_name = KERNEL_SPEC_TEMPLATE_NAMES[SPARK_OP]

            if self.image_name is None:
                self.image_name = f"{DEFAULT_SPARK_IMAGE_NAMES[self.language]}:{self._get_tag()}"
            if self.executor_image_name is None:
                self.executor_image_name = self.image_name

            # Spark operators use different provisioners and launchers
            if self.provisioner_name == PROVISIONER_NAME:
                self.provisioner_name = SPARK_OP_PROVISIONER_NAME
            if self.launcher_name == LAUNCHER_NAME:
                self.launcher_name = SPARK_OP_LAUNCHER_NAME
            self.launcher_dir_name = "operators"
        elif self.spark is True:
            # if kernel and display names are still defaulted, silently convert to lang default and append spark suffix
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[self.language] + SPARK_SUFFIX
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[self.language] + SPARK_DISPLAY_NAME_SUFFIX

            self.kernel_spec_dir_name = DEFAULT_KERNEL_NAMES[self.language] + SPARK_SUFFIX

            if self.image_name is None:
                self.image_name = f"{DEFAULT_SPARK_IMAGE_NAMES[self.language]}:{self._get_tag()}"
            if self.executor_image_name is None:
                self.executor_image_name = self.image_name
        else:
            # if kernel and display names are still defaulted, silently change to language defaults
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[self.language]
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[self.language]

            self.kernel_spec_dir_name = KERNEL_SPEC_TEMPLATE_NAMES[self.language]

            if self.image_name is None:
                self.image_name = f"{DEFAULT_IMAGE_NAMES[self.language]}:{self._get_tag()}"

            self.spark_init_mode = "none"
            if len(self.extra_spark_opts) > 0:
                self.log.warning(
                    "--extra_spark_opts will be ignored since --spark has not been specified."
                )
                self.extra_spark_opts = ""

        # sanitize kernel_name
        self.kernel_name = self.kernel_name.replace(" ", "_")

    @overrides
    def get_substitutions(self, install_dir) -> dict:
        substitutions = super().get_substitutions(install_dir)
        substitutions["image_name"] = self.image_name
        substitutions["executor_image_name"] = self.executor_image_name
        substitutions["provisioner_name"] = self.provisioner_name
        substitutions["launcher_name"] = self.launcher_name
        return substitutions


class K8sProvisionerApp(JupyterApp):
    """Application responsible for driving the creation of Kubernetes-based kernel specifications."""

    version = __version__
    name = "jupyter k8s-spec"
    description = (
        """Application used to create kernel specifications for use on Kubernetes clusters."""
    )
    subcommands = dict(
        {
            "install": (K8sSpecInstaller, K8sSpecInstaller.description.splitlines()[0]),
        }
    )
    aliases = {}
    flags = {}

    def start(self):
        super().start()

        if self.subapp is None:
            print(f"No subcommand specified. Must specify one of: {list(self.subcommands)}")
            print()
            self.print_description()
            self.print_subcommands()
            self.exit(1)


if __name__ == "__main__":
    K8sProvisionerApp.launch_instance()
